import time
from enum import Enum
from typing import Any, Dict, List, Optional, Union

import six
from pydantic import parse_obj_as

import prefect
from prefect import Task
from prefect.exceptions import PrefectException
from prefect.tasks.databricks.databricks_hook import DatabricksHook
from prefect.tasks.databricks.models import (
    AccessControlRequest,
    AccessControlRequestForGroup,
    AccessControlRequestForUser,
    GitSource,
    JobTaskSettings,
)
from prefect.utilities.tasks import defaults_from_attrs


def _deep_string_coerce(content, json_path="json"):
    """
    Coerces content or all values of content if it is a dict to a string. The
    function will throw if content contains non-string or non-numeric types.

    The reason why we have this function is because the `self.json` field must be a
    dict with only string values. This is because `render_template` will fail
    for numerical values.
    """
    if isinstance(content, six.string_types):
        return content
    elif isinstance(content, six.integer_types + (float,)):
        # Databricks can tolerate either numeric or string types in the API backend.
        return str(content)
    elif content is None:
        return content
    elif isinstance(content, (list, tuple)):
        return [
            _deep_string_coerce(content=item, json_path=f"{json_path}[{i}]")
            for i, item in enumerate(content)
        ]
    elif isinstance(content, dict):
        return {
            key: _deep_string_coerce(content=value, json_path=f"{json_path}[{key}]")
            for key, value in list(content.items())
        }
    elif isinstance(content, Enum):
        return str(content.value)
    raise ValueError(
        f"Type {type(content)} used for parameter {json_path} is not a number or a string"
    )


def _handle_databricks_task_execution(task, hook, log, submitted_run_id):
    """
    Handles the Databricks + Prefect lifecycle logic for a Databricks task

    Args:
        - task (prefect.Task) : Prefect task being handled
        - hook (prefect.tasks.databricks.databricks_hook.DatabricksHook): Databricks Hook
        - log (logger): Prefect logging instance
        - submitted_run_id (str): run ID returned after submitting or running Databricks job
    """

    log.info("Run submitted with run_id: %s", submitted_run_id)
    run_page_url = hook.get_run_page_url(submitted_run_id)

    log.info("Run submitted with config : %s", task.json)

    log.info("View run status, Spark UI, and logs at %s", run_page_url)
    while True:
        run_state = hook.get_run_state(submitted_run_id)
        if run_state.is_terminal:
            if run_state.is_successful:
                log.info("%s completed successfully.", task.name)
                log.info("View run status, Spark UI, and logs at %s", run_page_url)
                return
            else:
                error_message = "{t} failed with terminal state: {s}".format(
                    t=task.name, s=run_state
                )
                raise PrefectException(error_message)
        else:
            log.info("%s in run state: %s", task.name, run_state)
            log.info("View run status, Spark UI, and logs at %s", run_page_url)
            log.info("Sleeping for %s seconds.", task.polling_period_seconds)
            time.sleep(task.polling_period_seconds)


class DatabricksSubmitRun(Task):
    """
    Submits a Spark job run to Databricks using the
    [api/2.0/jobs/runs/submit](https://docs.databricks.com/api/latest/jobs.html#runs-submit)
    API endpoint.

    Currently the named parameters that `DatabricksSubmitRun` task supports are

    - `spark_jar_task`
    - `notebook_task`
    - `new_cluster`
    - `existing_cluster_id`
    - `libraries`
    - `run_name`
    - `timeout_seconds`

    Args:
        - databricks_conn_secret (dict, optional): Dictionary representation of the Databricks Connection
            String. Structure must be a string of valid JSON. To use token based authentication, provide
            the key `token` in the string for the connection and create the key `host`.
            `PREFECT__CONTEXT__SECRETS__DATABRICKS_CONNECTION_STRING=
            '{"host": "abcdef.xyz", "login": "ghijklmn", "password": "opqrst"}'`
            OR
            `PREFECT__CONTEXT__SECRETS__DATABRICKS_CONNECTION_STRING=
            '{"host": "abcdef.xyz", "token": "ghijklmn"}'`
            See documentation of the `DatabricksSubmitRun` Task to see how to pass in the connection
            string using `PrefectSecret`.
        - json (dict, optional): A JSON object containing API parameters which will be passed
            directly to the `api/2.0/jobs/runs/submit` endpoint. The other named parameters
            (i.e. `spark_jar_task`, `notebook_task`..) to this task will
            be merged with this json dictionary if they are provided.
            If there are conflicts during the merge, the named parameters will
            take precedence and override the top level json keys. (templated)
            For more information about templating see :ref:`jinja-templating`.
            https://docs.databricks.com/api/latest/jobs.html#runs-submit
        - spark_jar_task (dict, optional): The main class and parameters for the JAR task. Note that
            the actual JAR is specified in the `libraries`.
            *EITHER* `spark_jar_task` *OR* `notebook_task` should be specified.
            This field will be templated.
            https://docs.databricks.com/api/latest/jobs.html#jobssparkjartask
        - notebook_task (dict, optional): The notebook path and parameters for the notebook task.
            *EITHER* `spark_jar_task` *OR* `notebook_task` should be specified.
            This field will be templated.
            https://docs.databricks.com/api/latest/jobs.html#jobsnotebooktask
        - new_cluster (dict, optional): Specs for a new cluster on which this task will be run.
            *EITHER* `new_cluster` *OR* `existing_cluster_id` should be specified.
            This field will be templated.
            https://docs.databricks.com/api/latest/jobs.html#jobsclusterspecnewcluster
        - existing_cluster_id (str, optional): ID for existing cluster on which to run this task.
            *EITHER* `new_cluster` *OR* `existing_cluster_id` should be specified.
            This field will be templated.
        - libraries (list of dicts, optional): Libraries which this run will use.
            This field will be templated.
            https://docs.databricks.com/api/latest/libraries.html#managedlibrarieslibrary
        - run_name (str, optional): The run name used for this task.
            By default this will be set to the Prefect `task_id`. This `task_id` is a
            required parameter of the superclass `Task`.
            This field will be templated.
        - timeout_seconds (int, optional): The timeout for this run. By default a value of 0 is used
            which means to have no timeout.
            This field will be templated.
        - polling_period_seconds (int, optional): Controls the rate which we poll for the result of
            this run. By default the task will poll every 30 seconds.
        - databricks_retry_limit (int, optional): Amount of times retry if the Databricks backend is
            unreachable. Its value must be greater than or equal to 1.
        - databricks_retry_delay (float, optional): Number of seconds to wait between retries (it
            might be a floating point number).
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor

    Examples:
        There are two ways to instantiate this task.

        In the first way, you can take the JSON payload that you typically use
        to call the `api/2.0/jobs/runs/submit` endpoint and pass it directly
        to our `DatabricksSubmitRun` task through the `json` parameter.
        ```python
        from prefect import Flow
        from prefect.tasks.secrets import PrefectSecret
        from prefect.tasks.databricks import DatabricksSubmitRun

        json = {
            'new_cluster': {
                'spark_version': '10.4.x-scala2.12',
                'num_workers': 2,
                'node_type_id': "m4.large",
                'aws_attributes': {
                    'ebs_volume_type': "GENERAL_PURPOSE_SSD",
                    'ebs_volume_count': 3,
                    'ebs_volume_size': 100,
                }
            },
            'notebook_task': {
                'notebook_path': '/Users/andrew.h@prefect.io/records',
            },
        }

        with Flow("my flow") as flow:
            conn = PrefectSecret('DATABRICKS_CONNECTION_STRING')
            notebook_run = DatabricksSubmitRun(json=json)
            notebook_run(databricks_conn_secret=conn)
        ```

        Another way to accomplish the same thing is to use the named parameters
        of the `DatabricksSubmitRun` directly. Note that there is exactly
        one named parameter for each top level parameter in the `runs/submit`
        endpoint. In this method, your code would look like this:
        ```python
        from prefect import Flow
        from prefect.tasks.secrets import PrefectSecret
        from prefect.tasks.databricks import DatabricksSubmitRun

        new_cluster = {
            'spark_version': '10.4.x-scala2.12',
            'num_workers': 2,
            'node_type_id': "m4.large",
            'aws_attributes': {
                'ebs_volume_type': "GENERAL_PURPOSE_SSD",
                'ebs_volume_count': 3,
                'ebs_volume_size': 100,
            }
        }
        notebook_task = {
            'notebook_path': '/Users/prefect@example.com/PrepareData',
        }

        with Flow("my flow") as flow:
            conn = PrefectSecret('DATABRICKS_CONNECTION_STRING')
            notebook_run = DatabricksSubmitRun(
                new_cluster=new_cluster,
                notebook_task=notebook_task)
            notebook_run(databricks_conn_secret=conn)
        ```

        In the case where both the json parameter **AND** the named parameters
        are provided, they will be merged together. If there are conflicts during the merge,
        the named parameters will take precedence and override the top level `json` keys.
        This task requires a Databricks connection to be specified as a Prefect secret and can
        be passed to the task like so:
        ```
        from prefect import Flow
        from prefect.tasks.secrets import PrefectSecret
        from prefect.tasks.databricks import DatabricksSubmitRun

        with Flow("my flow") as flow:
            conn = PrefectSecret('DATABRICKS_CONNECTION_STRING')
            notebook_run = DatabricksSubmitRun(json=...)
            notebook_run(databricks_conn_secret=conn)
        ```
    """

    def __init__(
        self,
        databricks_conn_secret: Dict[str, str] = None,
        json: Dict[str, Any] = None,
        spark_jar_task: Optional[Dict[str, Any]] = None,
        notebook_task: Optional[Dict[str, Any]] = None,
        new_cluster: Optional[Dict[str, Any]] = None,
        existing_cluster_id: Optional[str] = None,
        libraries: Optional[List[Dict[str, Any]]] = None,
        run_name: Optional[str] = None,
        timeout_seconds: Optional[int] = None,
        polling_period_seconds: int = 30,
        databricks_retry_limit: int = 3,
        databricks_retry_delay: float = 1,
        **kwargs,
    ) -> None:

        self.databricks_conn_secret = databricks_conn_secret
        self.json = json or {}
        self.spark_jar_task = spark_jar_task
        self.notebook_task = notebook_task
        self.new_cluster = new_cluster
        self.existing_cluster_id = existing_cluster_id
        self.libraries = libraries
        self.run_name = run_name
        self.timeout_seconds = timeout_seconds
        self.polling_period_seconds = polling_period_seconds
        self.databricks_retry_limit = databricks_retry_limit
        self.databricks_retry_delay = databricks_retry_delay

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "databricks_conn_secret",
        "json",
        "spark_jar_task",
        "notebook_task",
        "new_cluster",
        "existing_cluster_id",
        "libraries",
        "run_name",
        "timeout_seconds",
        "polling_period_seconds",
        "databricks_retry_limit",
        "databricks_retry_delay",
    )
    def run(
        self,
        databricks_conn_secret: Dict[str, str] = None,
        json: Dict[str, Any] = None,
        spark_jar_task: Optional[Dict[str, Any]] = None,
        notebook_task: Optional[Dict[str, Any]] = None,
        new_cluster: Optional[Dict[str, Any]] = None,
        existing_cluster_id: Optional[str] = None,
        libraries: Optional[List[Dict[str, Any]]] = None,
        run_name: Optional[str] = None,
        timeout_seconds: Optional[int] = None,
        polling_period_seconds: int = 30,
        databricks_retry_limit: int = 3,
        databricks_retry_delay: float = 1,
    ) -> str:
        """
        Task run method.

        Args:
            - databricks_conn_secret (dict, optional): Dictionary representation of the Databricks
                Connection String. Structure must be a string of valid JSON. To use token based,
                authentication provide the key `token` in the string for the connection and
                create the key `host`.
                `PREFECT__CONTEXT__SECRETS__DATABRICKS_CONNECTION_STRING=
                '{"host": "abcdef.xyz", "login": "ghijklmn", "password": "opqrst"}'`
                OR
                `PREFECT__CONTEXT__SECRETS__DATABRICKS_CONNECTION_STRING=
                '{"host": "abcdef.xyz", "token": "ghijklmn"}'`
                See documentation of the `DatabricksSubmitRun` Task to see how to pass in the connection
                string using `PrefectSecret`.
            - json (dict, optional): A JSON object containing API parameters which will be passed
                directly to the `api/2.0/jobs/runs/submit` endpoint. The other named parameters
                (i.e. `spark_jar_task`, `notebook_task`..) to this task will
                be merged with this json dictionary if they are provided.
                If there are conflicts during the merge, the named parameters will
                take precedence and override the top level json keys. (templated)
                For more information about templating see :ref:`jinja-templating`.
                https://docs.databricks.com/api/latest/jobs.html#runs-submit
            - spark_jar_task (dict, optional): The main class and parameters for the JAR task. Note that
                the actual JAR is specified in the `libraries`.
                *EITHER* `spark_jar_task` *OR* `notebook_task` should be specified.
                This field will be templated.
                https://docs.databricks.com/api/latest/jobs.html#jobssparkjartask
            - notebook_task (dict, optional): The notebook path and parameters for the notebook task.
                *EITHER* `spark_jar_task` *OR* `notebook_task` should be specified.
                This field will be templated.
                https://docs.databricks.com/api/latest/jobs.html#jobsnotebooktask
            - new_cluster (dict, optional): Specs for a new cluster on which this task will be run.
                *EITHER* `new_cluster` *OR* `existing_cluster_id` should be specified.
                This field will be templated.
                https://docs.databricks.com/api/latest/jobs.html#jobsclusterspecnewcluster
            - existing_cluster_id (str, optional): ID for existing cluster on which to run this task.
                *EITHER* `new_cluster` *OR* `existing_cluster_id` should be specified.
                This field will be templated.
            - libraries (list of dicts, optional): Libraries which this run will use.
                This field will be templated.
                https://docs.databricks.com/api/latest/libraries.html#managedlibrarieslibrary
            - run_name (str, optional): The run name used for this task.
                By default this will be set to the Prefect `task_id`. This `task_id` is a
                required parameter of the superclass `Task`.
                This field will be templated.
            - timeout_seconds (int, optional): The timeout for this run. By default a value of 0 is used
                which means to have no timeout.
                This field will be templated.
            - polling_period_seconds (int, optional): Controls the rate which we poll for the result of
                this run. By default the task will poll every 30 seconds.
            - databricks_retry_limit (int, optional): Amount of times retry if the Databricks backend is
                unreachable. Its value must be greater than or equal to 1.
            - databricks_retry_delay (float, optional): Number of seconds to wait between retries (it
                might be a floating point number).

        Returns:
            - run_id (str): Run id of the submitted run
        """

        if databricks_conn_secret is None:
            raise ValueError(
                "A databricks connection string must be supplied as "
                "a dictionary or through Prefect Secrets"
            )
        elif not isinstance(databricks_conn_secret, dict):
            raise ValueError(
                "`databricks_conn_secret` must be supplied as a valid dictionary."
            )
        self.databricks_conn_secret = databricks_conn_secret

        if json:
            self.json = json
        if polling_period_seconds:
            self.polling_period_seconds = polling_period_seconds

        # Initialize Databricks Connections
        hook = DatabricksHook(
            databricks_conn_secret,
            retry_limit=databricks_retry_limit,
            retry_delay=databricks_retry_delay,
        )

        if spark_jar_task is not None:
            self.json["spark_jar_task"] = spark_jar_task
        if notebook_task is not None:
            self.json["notebook_task"] = notebook_task
        if new_cluster is not None:
            self.json["new_cluster"] = new_cluster
        if existing_cluster_id is not None:
            self.json["existing_cluster_id"] = existing_cluster_id
        if libraries is not None:
            self.json["libraries"] = libraries
        if run_name is not None:
            self.json["run_name"] = run_name
        if timeout_seconds is not None:
            self.json["timeout_seconds"] = timeout_seconds
        if "run_name" not in self.json:
            self.json["run_name"] = run_name or "Run Submitted by Prefect"

        # Validate the dictionary to a valid JSON object
        self.json = _deep_string_coerce(self.json)

        # Submit the job
        submitted_run_id = hook.submit_run(self.json)
        _handle_databricks_task_execution(self, hook, self.logger, submitted_run_id)

        return submitted_run_id


class DatabricksRunNow(Task):
    """
    Runs an existing Spark job run to Databricks using the
    [api/2.1/jobs/run-now](https://docs.databricks.com/api/latest/jobs.html#run-now)
    API endpoint.

    Currently the named parameters that `DatabricksRunNow` task supports are

    - `job_id`
    - `json`
    - `notebook_params`
    - `python_params`
    - `spark_submit_params`
    - `jar_params`

    Args:
        - databricks_conn_secret (dict, optional): Dictionary representation of the Databricks Connection
            String. Structure must be a string of valid JSON. To use token based authentication, provide
            the key `token` in the string for the connection and create the key `host`.
            `PREFECT__CONTEXT__SECRETS__DATABRICKS_CONNECTION_STRING=
            '{"host": "abcdef.xyz", "login": "ghijklmn", "password": "opqrst"}'`
            OR
            `PREFECT__CONTEXT__SECRETS__DATABRICKS_CONNECTION_STRING=
            '{"host": "abcdef.xyz", "token": "ghijklmn"}'`
            See documentation of the `DatabricksSubmitRun` Task to see how to pass in the connection
            string using `PrefectSecret`.
        - job_id (str, optional): The job_id of the existing Databricks job.
            https://docs.databricks.com/api/latest/jobs.html#run-now
        - json (dict, optional): A JSON object containing API parameters which will be passed
            directly to the `api/2.0/jobs/run-now` endpoint. The other named parameters
            (i.e. `notebook_params`, `spark_submit_params`..) to this operator will
            be merged with this json dictionary if they are provided.
            If there are conflicts during the merge, the named parameters will
            take precedence and override the top level json keys. (templated)
            https://docs.databricks.com/api/latest/jobs.html#run-now
        - notebook_params (dict, optional): A dict from keys to values for jobs with notebook task,
            e.g. "notebook_params": {"name": "john doe", "age":  "35"}.
            The map is passed to the notebook and will be accessible through the
            dbutils.widgets.get function. See Widgets for more information.
            If not specified upon run-now, the triggered run will use the
            job’s base parameters. notebook_params cannot be
            specified in conjunction with jar_params. The json representation
            of this field (i.e. {"notebook_params":{"name":"john doe","age":"35"}})
            cannot exceed 10,000 bytes.
            https://docs.databricks.com/user-guide/notebooks/widgets.html
        - python_params (list[str], optional): A list of parameters for jobs with python tasks,
            e.g. "python_params": ["john doe", "35"].
            The parameters will be passed to python file as command line parameters.
            If specified upon run-now, it would overwrite the parameters specified in
            job setting.
            The json representation of this field (i.e. {"python_params":["john doe","35"]})
            cannot exceed 10,000 bytes.
            https://docs.databricks.com/api/latest/jobs.html#run-now
        - spark_submit_params (list[str], optional): A list of parameters for jobs with spark submit
            task, e.g. "spark_submit_params": ["--class", "org.apache.spark.examples.SparkPi"].
            The parameters will be passed to spark-submit script as command line parameters.
            If specified upon run-now, it would overwrite the parameters specified
            in job setting.
            The json representation of this field cannot exceed 10,000 bytes.
            https://docs.databricks.com/api/latest/jobs.html#run-now
        - jar_params (list[str], optional): A list of parameters for jobs with JAR tasks,
            e.g. "jar_params": ["john doe", "35"]. The parameters will be used to invoke the main
            function of the main class specified in the Spark JAR task. If not specified upon
            run-now, it will default to an empty list. jar_params cannot be specified in conjunction
            with notebook_params. The JSON representation of this field (i.e.
            {"jar_params":["john doe","35"]}) cannot exceed 10,000 bytes.
            https://docs.databricks.com/api/latest/jobs.html#run-now
        - timeout_seconds (int, optional): The timeout for this run. By default a value of 0 is used
            which means to have no timeout.
            This field will be templated.
        - polling_period_seconds (int, optional): Controls the rate which we poll for the result of
            this run. By default the task will poll every 30 seconds.
        - databricks_retry_limit (int, optional): Amount of times retry if the Databricks backend is
            unreachable. Its value must be greater than or equal to 1.
        - databricks_retry_delay (float, optional): Number of seconds to wait between retries (it
            might be a floating point number).
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor

    Examples:
        There are two ways to instantiate this task.

        In the first way, you can take the JSON payload that you typically use
        to call the `api/2.1/jobs/run-now` endpoint and pass it directly
        to our `DatabricksRunNow` task through the `json` parameter.
        ```python
        from prefect import Flow
        from prefect.tasks.secrets import PrefectSecret
        from prefect.tasks.databricks import DatabricksRunNow

        json = {
            "job_id": 42,
            "notebook_params": {
                "dry-run": "true",
                "oldest-time-to-consider": "1457570074236"
            }
        }

        with Flow("my flow") as flow:
            conn = PrefectSecret('DATABRICKS_CONNECTION_STRING')
            notebook_run = DatabricksRunNow(json=json)
            notebook_run(databricks_conn_secret=conn)
        ```

        Another way to accomplish the same thing is to use the named parameters
        of the `DatabricksRunNow` task directly. Note that there is exactly
        one named parameter for each top level parameter in the `run-now`
        endpoint. In this method, your code would look like this:
        ```python
        from prefect import Flow
        from prefect.tasks.secrets import PrefectSecret
        from prefect.tasks.databricks import DatabricksRunNow

        job_id = 42

        notebook_params = {
            "dry-run": "true",
            "oldest-time-to-consider": "1457570074236"
        }

        python_params = ["douglas adams", "42"]
        spark_submit_params = ["--class", "org.apache.spark.examples.SparkPi"]
        jar_params = ["john doe","35"]

        with Flow("my flow') as flow:
            conn = PrefectSecret('DATABRICKS_CONNECTION_STRING')
            notebook_run = DatabricksRunNow(
                notebook_params=notebook_params,
                python_params=python_params,
                spark_submit_params=spark_submit_params,
                jar_params=jar_params
            )
            notebook_run(databricks_conn_secret=conn)
        ```

        In the case where both the json parameter **AND** the named parameters
        are provided, they will be merged together. If there are conflicts during the merge,
        the named parameters will take precedence and override the top level `json` keys.

        This task requires a Databricks connection to be specified as a Prefect secret and can
        be passed to the task like so:
        ```
        from prefect import Flow
        from prefect.tasks.secrets import PrefectSecret
        from prefect.tasks.databricks import DatabricksRunNow

        with Flow("my flow") as flow:
            conn = PrefectSecret('DATABRICKS_CONNECTION_STRING')
            notebook_run = DatabricksRunNow(json=...)
            notebook_run(databricks_conn_secret=conn)
        ```

    """

    def __init__(
        self,
        databricks_conn_secret: Dict[str, str] = None,
        job_id: str = None,
        json: Dict[str, Any] = None,
        notebook_params: Optional[Dict[str, Any]] = None,
        python_params: Optional[List[str]] = None,
        spark_submit_params: Optional[List[str]] = None,
        jar_params: Optional[List[str]] = None,
        polling_period_seconds: int = 30,
        databricks_retry_limit: int = 3,
        databricks_retry_delay: float = 1,
        **kwargs,
    ) -> None:

        self.databricks_conn_secret = databricks_conn_secret
        self.json = json or {}
        self.job_id = job_id
        self.notebook_params = notebook_params
        self.python_params = python_params
        self.spark_submit_params = spark_submit_params
        self.jar_params = jar_params
        self.polling_period_seconds = polling_period_seconds
        self.databricks_retry_limit = databricks_retry_limit
        self.databricks_retry_delay = databricks_retry_delay

        super().__init__(**kwargs)

    @staticmethod
    def _get_hook(
        databricks_conn_secret, databricks_retry_limit, databricks_retry_delay
    ):
        # this is a function so it's easier to mock in testing
        return DatabricksHook(
            databricks_conn_secret,
            retry_limit=databricks_retry_limit,
            retry_delay=databricks_retry_delay,
        )

    @defaults_from_attrs(
        "databricks_conn_secret",
        "job_id",
        "json",
        "notebook_params",
        "python_params",
        "spark_submit_params",
        "jar_params",
        "polling_period_seconds",
        "databricks_retry_limit",
        "databricks_retry_delay",
    )
    def run(
        self,
        databricks_conn_secret: Dict[str, str] = None,
        job_id: str = None,
        json: Dict[str, Any] = None,
        notebook_params: Optional[Dict[str, Any]] = None,
        python_params: Optional[List[str]] = None,
        spark_submit_params: Optional[List[str]] = None,
        jar_params: Optional[List[str]] = None,
        polling_period_seconds: int = 30,
        databricks_retry_limit: int = 3,
        databricks_retry_delay: float = 1,
    ) -> str:
        """
        Task run method.

        Args:

            - databricks_conn_secret (dict, optional): Dictionary representation of the Databricks
                Connection String. Structure must be a string of valid JSON. To use token based
                authentication, provide the key `token` in the string for the connection and create the
                key `host`.
                `PREFECT__CONTEXT__SECRETS__DATABRICKS_CONNECTION_STRING=
                '{"host": "abcdef.xyz", "login": "ghijklmn", "password": "opqrst"}'`
                OR
                `PREFECT__CONTEXT__SECRETS__DATABRICKS_CONNECTION_STRING=
                '{"host": "abcdef.xyz", "token": "ghijklmn"}'`
                See documentation of the `DatabricksSubmitRun` Task to see how to pass in the connection
                string using `PrefectSecret`.
            - job_id (str, optional): The job_id of the existing Databricks job.
                https://docs.databricks.com/api/latest/jobs.html#run-now
            - json (dict, optional): A JSON object containing API parameters which will be passed
                directly to the `api/2.0/jobs/run-now` endpoint. The other named parameters
                (i.e. `notebook_params`, `spark_submit_params`..) to this operator will
                be merged with this json dictionary if they are provided.
                If there are conflicts during the merge, the named parameters will
                take precedence and override the top level json keys. (templated)
                https://docs.databricks.com/api/latest/jobs.html#run-now
            - notebook_params (dict, optional): A dict from keys to values for jobs with notebook task,
                e.g. "notebook_params": {"name": "john doe", "age":  "35"}.
                The map is passed to the notebook and will be accessible through the
                dbutils.widgets.get function. See Widgets for more information.
                If not specified upon run-now, the triggered run will use the
                job’s base parameters. notebook_params cannot be
                specified in conjunction with jar_params. The json representation
                of this field (i.e. {"notebook_params":{"name":"john doe","age":"35"}})
                cannot exceed 10,000 bytes.
                https://docs.databricks.com/user-guide/notebooks/widgets.html
            - python_params (list[str], optional): A list of parameters for jobs with python tasks,
                e.g. "python_params": ["john doe", "35"].
                The parameters will be passed to python file as command line parameters.
                If specified upon run-now, it would overwrite the parameters specified in
                job setting.
                The json representation of this field (i.e. {"python_params":["john doe","35"]})
                cannot exceed 10,000 bytes.
                https://docs.databricks.com/api/latest/jobs.html#run-now
            - spark_submit_params (list[str], optional): A list of parameters for jobs with spark submit
                task, e.g. "spark_submit_params": ["--class", "org.apache.spark.examples.SparkPi"].
                The parameters will be passed to spark-submit script as command line parameters.
                If specified upon run-now, it would overwrite the parameters specified
                in job setting.
                The json representation of this field cannot exceed 10,000 bytes.
                https://docs.databricks.com/api/latest/jobs.html#run-now
            - jar_params (list[str], optional): A list of parameters for jobs with JAR tasks,
                e.g. "jar_params": ["john doe", "35"]. The parameters will be used to invoke the main
                function of the main class specified in the Spark JAR task. If not specified upon
                run-now, it will default to an empty list. jar_params cannot be specified in conjunction
                with notebook_params. The JSON representation of this field (i.e.
                {"jar_params":["john doe","35"]}) cannot exceed 10,000 bytes.
                https://docs.databricks.com/api/latest/jobs.html#run-now
            - polling_period_seconds (int, optional): Controls the rate which we poll for the result of
                this run. By default the task will poll every 30 seconds.
            - databricks_retry_limit (int, optional): Amount of times retry if the Databricks backend is
                unreachable. Its value must be greater than or equal to 1.
            - databricks_retry_delay (float, optional): Number of seconds to wait between retries (it
                might be a floating point number).

        Returns:
            - run_id (str): Run id of the submitted run
        """
        if databricks_conn_secret is None:
            raise ValueError(
                "A databricks connection string must be supplied as "
                "a dictionary or through Prefect Secrets"
            )
        elif not isinstance(databricks_conn_secret, dict):
            raise ValueError(
                "`databricks_conn_secret` must be supplied as a valid dictionary."
            )
        self.databricks_conn_secret = databricks_conn_secret

        # Initialize Databricks Connections
        hook = self._get_hook(
            databricks_conn_secret, databricks_retry_limit, databricks_retry_delay
        )

        run_now_json = json or {}
        # Necessary be cause `_handle_databricks_task_execution` reads
        # `polling_periods_seconds` off of the task instance
        if polling_period_seconds:
            self.polling_period_seconds = polling_period_seconds

        if job_id is not None:
            run_now_json["job_id"] = job_id
        if notebook_params is not None:
            merged = run_now_json.setdefault("notebook_params", {})
            merged.update(notebook_params)
            run_now_json["notebook_params"] = merged
        if python_params is not None:
            run_now_json["python_params"] = python_params
        if spark_submit_params is not None:
            run_now_json["spark_submit_params"] = spark_submit_params
        if jar_params is not None:
            run_now_json["jar_params"] = jar_params

        # Validate the dictionary to a valid JSON object
        self.json = _deep_string_coerce(run_now_json)

        # Submit the job
        submitted_run_id = hook.run_now(self.json)
        _handle_databricks_task_execution(self, hook, self.logger, submitted_run_id)

        return submitted_run_id


class DatabricksSubmitMultitaskRun(Task):
    """
    Creates and triggers a one-time run via the Databricks submit run API endpoint. Supports
    the execution of multiple Databricks tasks within the Databricks job run. Note: Databricks
    tasks are distinct from Prefect tasks. All tasks configured will run as a single Prefect task.

    For more information about the arguments of this task, refer to the [Databricks submit run API
    documentation](https://docs.databricks.com/dev-tools/api/latest/jobs.html#operation/JobsRunsSubmit)

    Args:
        - databricks_conn_secret (dict, optional): Dictionary representation of the Databricks Connection
            String. Structure must be a string of valid JSON. To use token based authentication, provide
            the key `token` in the string for the connection and create the key `host`.
            `PREFECT__CONTEXT__SECRETS__DATABRICKS_CONNECTION_STRING=
            '{"host": "abcdef.xyz", "login": "ghijklmn", "password": "opqrst"}'`
            OR
            `PREFECT__CONTEXT__SECRETS__DATABRICKS_CONNECTION_STRING=
            '{"host": "abcdef.xyz", "token": "ghijklmn"}'`
        - tasks (List[JobTaskSettings]): A list containing the Databricks task configuration. Should
            contain configuration for at least one task.
        - timeout_seconds (int, optional):  An optional timeout applied to each run of this job.
            The default behavior is to have no timeout.
        - run_name (str, optional): An optional name for the run.
            The default value is "Job run created by Prefect flow run {flow_run_name}".
        - idempotency_token (str, optional): An optional token that can be used to guarantee
            the idempotency of job run requests. Defaults to the Python object id.
        - access_control_list (List[AccessControlRequest]): List of permissions to set on the job.
        - polling_period_seconds (int, optional): Controls the rate which we poll for the result of
            this run. By default the task will poll every 30 seconds.
        - databricks_retry_limit (int, optional): Amount of times retry if the Databricks backend is
            unreachable. Its value must be greater than or equal to 1.
        - databricks_retry_delay (float, optional): Number of seconds to wait between retries (it
            might be a floating point number).
        - git_source (GitSource): A git source for the source code of the jobs (see
            https://databricks.com/blog/2022/06/21/build-reliable-production-data-and-ml-pipelines-with-git-support-for-databricks-workflows.html)
        - **kwargs (dict, optional): Additional keyword arguments to pass to the
            Task constructor

    Examples:
        Trigger an ad-hoc multitask run

        ```
        from prefect import Flow
        from prefect.tasks.databricks import DatabricksSubmitMultitaskRun
        from prefect.tasks.databricks.models import (
            AccessControlRequestForUser,
            AutoScale,
            AwsAttributes,
            AwsAvailability,
            CanManage,
            JobTaskSettings,
            Library,
            NewCluster,
            NotebookTask,
            SparkJarTask,
            TaskDependency,
        )

        submit_multitask_run = DatabricksSubmitMultitaskRun(
            tasks=[
                JobTaskSettings(
                    task_key="Sessionize",
                    description="Extracts session data from events",
                    existing_cluster_id="0923-164208-meows279",
                    spark_jar_task=SparkJarTask(
                        main_class_name="com.databricks.Sessionize",
                        parameters=["--data", "dbfs:/path/to/data.json"],
                    ),
                    libraries=[Library(jar="dbfs:/mnt/databricks/Sessionize.jar")],
                    timeout_seconds=86400,
                ),
                JobTaskSettings(
                    task_key="Orders_Ingest",
                    description="Ingests order data",
                    existing_cluster_id="0923-164208-meows279",
                    spark_jar_task=SparkJarTask(
                        main_class_name="com.databricks.OrdersIngest",
                        parameters=["--data", "dbfs:/path/to/order-data.json"],
                    ),
                    libraries=[Library(jar="dbfs:/mnt/databricks/OrderIngest.jar")],
                    timeout_seconds=86400,
                ),
                JobTaskSettings(
                    task_key="Match",
                    description="Matches orders with user sessions",
                    depends_on=[
                        TaskDependency(task_key="Orders_Ingest"),
                        TaskDependency(task_key="Sessionize"),
                    ],
                    new_cluster=NewCluster(
                        spark_version="10.4.x-scala2.12",
                        node_type_id="m4.large",
                        spark_conf={"spark.speculation": True},
                        aws_attributes=AwsAttributes(
                            availability=AwsAvailability.SPOT,
                            zone_id="us-west-2a",
                            ebs_volume_type="GENERAL_PURPOSE_SSD",
                            ebs_volume_count=3,
                            ebs_volume_size=100,
                        ),
                        autoscale=AutoScale(min_workers=1, max_workers=2),
                    ),
                    notebook_task=NotebookTask(
                        notebook_path="/Users/user.name@databricks.com/Match",
                        base_parameters={"name": "John Doe", "age": "35"},
                    ),
                    timeout_seconds=86400,
                ),
            ],
            run_name="A multitask job run",
            timeout_seconds=86400,
            access_control_list=[
                AccessControlRequestForUser(
                    user_name="jsmith@example.com", permission_level=CanManage.CAN_MANAGE
                )
            ],
        )


        with Flow("my flow") as f:
            conn = PrefectSecret('DATABRICKS_CONNECTION_STRING')
            submit_multitask_run(databricks_conn_secret=conn)
        ```
    """

    def __init__(
        self,
        databricks_conn_secret: Dict[str, str] = None,
        tasks: List[JobTaskSettings] = None,
        run_name: str = None,
        timeout_seconds: int = None,
        idempotency_token: str = None,
        access_control_list: List[
            Union[AccessControlRequestForUser, AccessControlRequestForGroup]
        ] = None,
        polling_period_seconds: int = 30,
        databricks_retry_limit: int = 3,
        databricks_retry_delay: float = 1,
        git_source: GitSource = None,
        **kwargs,
    ):
        self.databricks_conn_secret = databricks_conn_secret
        self.tasks = tasks
        self.run_name = run_name
        self.timeout_seconds = timeout_seconds
        self.idempotency_token = idempotency_token
        self.access_control_list = access_control_list
        self.polling_period_seconds = polling_period_seconds
        self.databricks_retry_limit = databricks_retry_limit
        self.databricks_retry_delay = databricks_retry_delay
        self.git_source = git_source
        super().__init__(**kwargs)

    @staticmethod
    def convert_dict_to_kwargs(input: Dict[str, Any]) -> Dict[str, Any]:
        """
        Method to convert a dict that matches the structure of the Databricks API call into the required
        object types for the task input

        Args:
            - input (Dict): A dictionary representing the input to the task

        Returns:
            - A dictionary with values that match the input types of the class

        Example:
        Use a JSON-like dict as input

        ```
        from prefect import Flow
        from prefect.tasks.databricks import DatabricksSubmitMultitaskRun

        submit_multitask_run = DatabricksSubmitMultitaskRun()

        databricks_kwargs = DatabricksSubmitMultitaskRun.convert_dict_to_kwargs({
            "tasks": [
                {
                    "task_key": "Sessionize",
                    "description": "Extracts session data from events",
                    "depends_on": [],
                    "existing_cluster_id": "0923-164208-meows279",
                    "spark_jar_task": {
                        "main_class_name": "com.databricks.Sessionize",
                        "parameters": ["--data", "dbfs:/path/to/data.json"],
                    },
                    "libraries": [{"jar": "dbfs:/mnt/databricks/Sessionize.jar"}],
                    "timeout_seconds": 86400,
                },
                {
                    "task_key": "Orders_Ingest",
                    "description": "Ingests order data",
                    "depends_on": [],
                    "existing_cluster_id": "0923-164208-meows279",
                    "spark_jar_task": {
                        "main_class_name": "com.databricks.OrdersIngest",
                        "parameters": ["--data", "dbfs:/path/to/order-data.json"],
                    },
                    "libraries": [{"jar": "dbfs:/mnt/databricks/OrderIngest.jar"}],
                    "timeout_seconds": 86400,
                },
                {
                    "task_key": "Match",
                    "description": "Matches orders with user sessions",
                    "depends_on": [
                        {"task_key": "Orders_Ingest"},
                        {"task_key": "Sessionize"},
                    ],
                    "new_cluster": {
                        "spark_version": "7.3.x-scala2.12",
                        "node_type_id": "i3.xlarge",
                        "spark_conf": {"spark.speculation": True},
                        "aws_attributes": {
                            "availability": "SPOT",
                            "zone_id": "us-west-2a",
                        },
                        "autoscale": {"min_workers": 2, "max_workers": 16},
                    },
                    "notebook_task": {
                        "notebook_path": "/Users/user.name@databricks.com/Match",
                        "base_parameters": {"name": "John Doe", "age": "35"},
                    },
                    "timeout_seconds": 86400,
                },
            ],
            "run_name": "A multitask job run",
            "timeout_seconds": 86400,
            "access_control_list": [
                {
                    "user_name": "jsmith@example.com",
                    "permission_level": "CAN_MANAGE",
                }
            ],
        })

        with Flow("my flow") as f:
            conn = PrefectSecret('DATABRICKS_CONNECTION_STRING')
            submit_multitask_run(**databricks_kwargs, databricks_conn_secret=conn)
        ```
        """
        kwargs = {**input, "tasks": parse_obj_as(List[JobTaskSettings], input["tasks"])}

        if input.get("access_control_list"):
            kwargs["access_control_list"] = parse_obj_as(
                List[Union[AccessControlRequestForUser, AccessControlRequestForGroup]],
                input["access_control_list"],
            )
        if input.get("git_source"):
            kwargs["git_source"] = parse_obj_as(GitSource, input["git_source"])
        return kwargs

    @defaults_from_attrs(
        "databricks_conn_secret",
        "tasks",
        "run_name",
        "timeout_seconds",
        "idempotency_token",
        "access_control_list",
        "polling_period_seconds",
        "databricks_retry_limit",
        "databricks_retry_delay",
        "git_source",
    )
    def run(
        self,
        databricks_conn_secret: dict = None,
        tasks: List[JobTaskSettings] = None,
        run_name: str = None,
        timeout_seconds: int = None,
        idempotency_token: str = None,
        access_control_list: List[AccessControlRequest] = None,
        polling_period_seconds: int = None,
        databricks_retry_limit: int = None,
        databricks_retry_delay: float = None,
        git_source: GitSource = None,
    ):
        """
        Task run method. Any values passed here will overwrite the values used when initializing the
        task.

        Args:
            - databricks_conn_secret (dict, optional): Dictionary representation of the Databricks
                Connection String. Structure must be a string of valid JSON. To use token based
                authentication, provide the key `token` in the string for the connection and create
                the key `host`. `PREFECT__CONTEXT__SECRETS__DATABRICKS_CONNECTION_STRING=
                '{"host": "abcdef.xyz", "login": "ghijklmn", "password": "opqrst"}'`
                OR
                `PREFECT__CONTEXT__SECRETS__DATABRICKS_CONNECTION_STRING=
                '{"host": "abcdef.xyz", "token": "ghijklmn"}'`
            - tasks (List[JobTaskSettings]):" A list containing the Databricks task configuration. Should
                contain configuration for at least one task.
            - timeout_seconds (int, optional):  An optional timeout applied to each run of this job.
                The default behavior is to have no timeout.
            - run_name (str, optional): An optional name for the run.
                The default value is "Job run created by Prefect flow run {flow_run_name}".
            - idempotency_token (str, optional): An optional token that can be used to guarantee
                the idempotency of job run requests. Defaults to the Python object id.
            - access_control_list (List[AccessControlRequest]): List of permissions to set on the job.
            - polling_period_seconds (int, optional): Controls the rate which we poll for the result of
                this run. By default the task will poll every 30 seconds.
            - databricks_retry_limit (int, optional): Amount of times retry if the Databricks backend is
                unreachable. Its value must be greater than or equal to 1.
            - databricks_retry_delay (float, optional): Number of seconds to wait between retries (it
                might be a floating point number).
            - git_source (GitSource): A git source for the source code of the jobs (see
                https://databricks.com/blog/2022/06/21/build-reliable-production-data-and-ml-pipelines-with-git-support-for-databricks-workflows.html)

        Returns:
            - run_id (str): Run id of the submitted run

        """
        if databricks_conn_secret is None or not isinstance(
            databricks_conn_secret, dict
        ):
            raise ValueError(
                "Databricks connection info must be supplied as a dictionary."
            )
        if tasks is None or len(tasks) < 1:
            raise ValueError("Please supply at least one Databricks task to be run.")
        run_name = (
            run_name
            or f"Job run created by Prefect flow run {prefect.context.flow_run_name}"
        )
        # Ensures that multiple job runs are not created on retries
        idempotency_token = idempotency_token or id(self)

        # Set polling_period_seconds on task because _handle_databricks_task_execution expects it
        if polling_period_seconds:
            self.polling_period_seconds = polling_period_seconds

        databricks_client = DatabricksHook(
            databricks_conn_secret,
            retry_limit=databricks_retry_limit,
            retry_delay=databricks_retry_delay,
        )
        if git_source:
            git_source_json = git_source.dict()
        else:
            git_source_json = None
        # Set json on task instance because _handle_databricks_task_execution expects it
        self.json = _deep_string_coerce(
            dict(
                tasks=[task.dict() for task in tasks],
                run_name=run_name,
                timeout_seconds=timeout_seconds,
                idempotency_token=idempotency_token,
                access_control_list=[
                    entry.dict() for entry in access_control_list or []
                ],
                git_source=git_source_json,
            )
        )

        submitted_run_id = databricks_client.submit_multi_task_run(self.json)
        _handle_databricks_task_execution(
            self, databricks_client, self.logger, submitted_run_id
        )

        return submitted_run_id



"""
Tasks for interacting with Prometheus. The main task are using pushgateway 
"""
try:
    from prefect.tasks.prometheus.pushgateway import (
        PushGaugeToGateway,
        PushAddGaugeToGateway,
    )
except ImportError as err:
    raise ImportError(
        'Using `prefect.tasks.prometheus` requires Prefect to be installed with the "prometheus" extra.'
    ) from err

__all__ = ["PushAddGaugeToGateway", "PushGaugeToGateway"]



from collections import defaultdict
from datetime import datetime
import os

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs
from prefect.engine.signals import FAIL

import requests
import time
from typing import List, Union


class ZendeskTicketsIncrementalExportTask(Task):
    """
    This task can be used to perform an incremental export of tickets from Zendesk.
    More info at
    https://developer.zendesk.com/api-reference/ticketing/ticket-management/incremental_exports/#incremental-ticket-export.

    Args:
        - subdomain (str, optional): The Zendesk subdomain to use to export tickets.
        - email_address (str, optional): The email address to use to authenticate on Zendesk.
        - api_token (str, optional): The API token to use to athenticate on Zendesk
            If passed, it will take precedence over `api_token_env_var`.
        - api_token_env_var (str, optional): The name of the env var which contains the
            API token to use to authenticate on Zendesk.
        - start_time (int, datetime, optional): The start time to use to export tickets.
            Can be passed as an epoch timestamp or a `datetime` object.
        - cursor (str, optional): The cursor to use to export tickets.
            If passed, it will take precedence over `start_time`.
        - exclude_deleted: (bool, optional): Whether to exclude deleted tickets or not.
            Default to `False`.
        - include_entities: (str, list, optional): Optional list of entities to side load.
          More info at
          https://developer.zendesk.com/documentation/ticketing/using-the-zendesk-api/side_loading/.
        - **kwargs (dict, optional): Additional keyword arguments to pass to the
            Task constructor.
    """

    _ZENDESK_API_BASE_URL = "https://{subdomain}.zendesk.com/api/v2"

    def __init__(
        self,
        subdomain: str = None,
        email_address: str = None,
        api_token: str = None,
        api_token_env_var: str = None,
        start_time: Union[int, datetime] = None,
        cursor: str = None,
        exclude_deleted: bool = False,
        include_entities: List[str] = None,
        **kwargs,
    ):
        self.subdomain = subdomain
        self.email_address = email_address
        self.api_token = api_token
        self.api_token_env_var = api_token_env_var
        self.start_time = start_time
        self.cursor = cursor
        self.exclude_deleted = exclude_deleted
        self.include_entities = include_entities
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "subdomain",
        "email_address",
        "api_token",
        "api_token_env_var",
        "start_time",
        "cursor",
        "exclude_deleted",
        "include_entities",
    )
    def run(
        self,
        subdomain: str = None,
        email_address: str = None,
        api_token: str = None,
        api_token_env_var: str = None,
        start_time: Union[int, datetime] = None,
        cursor: str = None,
        exclude_deleted: bool = None,
        include_entities: List[str] = None,
    ):
        """
        Task run method to perform an incremental export of tickets from Zendesk.
        Args:
            - subdomain (str, optional): The Zendesk subdomain to use to export tickets.
            - email_address (str, optional): The email address to use to authenticate on Zendesk.
            - api_token (str, optional): The API token to use to athenticate on Zendesk
                If passed, it will take precedence over `api_token_env_var`.
            - api_token_env_var (str, optional): The name of the env var which contains the
                API token to use to authenticate on Zendesk.
            - start_time (int, datetime, optional): The start time to use to export tickets.
                Can be passed as an epoch timestamp or a `datetime` object.
            - cursor (str, optional): The cursor to use to export tickets.
                If passed, it will take precedence over `start_time`.
            - exclude_deleted: (bool, optional): Whether to exclude deleted tickets or not.
                Default to `False`.
            - include_entities: (str, list, optional): Optional list of entities to side load.
                More info at
                https://developer.zendesk.com/documentation/ticketing/using-the-zendesk-api/side_loading/.

        Raises:
            - `ValueError` if both `api_token` and `api_token_env_var` are missing.
            - `ValueError` if `api_token` is missing and `api_token_env_var` cannot be found.
            - `ValueError` if `subdomain` is missing.
            - `ValueError` if `email_address` is missing.
            - `ValueError` if both `start_time` and `cursor` are missing.
            - `prefect.engine.signals.FAIL` if the Zendesk API call fails.

        Returns:
            - A `dict` containing the list of tickets and, optionally, the included
              entities.
        """
        if not api_token and not api_token_env_var:
            raise ValueError("Both `api_token` and `api_token_env_var` are missing.")

        if not api_token and api_token_env_var not in os.environ:
            raise ValueError(
                "`api_token` is missing and `api_token_env_var` not found."
            )

        token = None
        if api_token:
            token = api_token
        elif api_token_env_var:
            token = os.environ[api_token_env_var]

        if not subdomain:
            raise ValueError("`subdomain` is missing.")

        if not email_address:
            raise ValueError("`email_address` is missing.")

        if not start_time and not cursor:
            raise ValueError("Both `start_time` and `cursor` are missing.")

        base_url = self._ZENDESK_API_BASE_URL.format(subdomain=subdomain)
        export_url = f"{base_url}/incremental/tickets/cursor.json"

        if cursor:
            self.logger.debug("Got cursor")
            export_url = f"{export_url}?cursor={cursor}"

        elif start_time:
            self.logger.debug("Got start_time")
            start_datetime = (
                start_time
                if isinstance(start_time, int)
                else int(start_time.timestamp())
            )
            export_url = f"{export_url}?start_time={start_datetime}"

        if exclude_deleted:
            export_url = f"{export_url}&exclude_deleted=true"

        if include_entities:
            if isinstance(include_entities, str):
                include_entities_str = include_entities
            elif isinstance(include_entities, list):
                include_entities_str = ",".join(list(set(include_entities)))

            export_url = f"{export_url}&include={include_entities_str}"

        session = requests.Session()
        session.auth = f"{email_address}/token", token

        end_of_stream = False

        tickets = defaultdict(list)

        while not end_of_stream:
            with session.get(export_url) as response:
                self.logger.debug(f"Export URL is: {export_url}")

                if response.status_code == 429:
                    retry_after_seconds = int(response.headers["retry-after"])
                    self.logger.warning(
                        f"""
                        API rate limit reached!
                        Waiting for {retry_after_seconds} seconds before retrying.
                        """
                    )
                    time.sleep(retry_after_seconds + 1)
                    continue

                elif response.status_code != 200:
                    msg = f"""
                    Zendesk API call failed!
                    Status: {response.status_code}
                    Reason: {response.reason}
                    """
                    raise FAIL(message=msg)

                content = response.json()

                tickets["tickets"].extend(content["tickets"])

                if include_entities:
                    for include_entity in list(set(include_entities)):
                        if include_entity in content.keys():
                            tickets[include_entity].extend(content[include_entity])

                end_of_stream = content["end_of_stream"]
                export_url = content["after_url"]
                cursor = content["after_cursor"]

            if not end_of_stream:
                # Try to avoid the rate limit: 10 requests per minute
                time.sleep(0.1)

        return tickets



import os
import socket
import sys
from subprocess import STDOUT, Popen, DEVNULL
from typing import Iterable, List

from prefect import config
from prefect.agent import Agent
from prefect.storage import Docker
from prefect.run_configs import LocalRun
from prefect.serialization.storage import StorageSchema
from prefect.utilities.graphql import GraphQLResult


class LocalAgent(Agent):
    """
    Agent which deploys flow runs locally as subprocesses. There are a range of kwarg
    options to control information which may be provided to these subprocesses.

    Optional import paths may be specified to append dependency modules to the PATH:
    ```
    prefect agent local start --import-path "/usr/local/my_module" --import-path "~/other_module"

    # Now the local scripts/packages my_module and other_module will be importable in
    # the flow's subprocess
    ```

    Environment variables may be set on the agent to be provided to each flow run's subprocess:
    ```
    prefect agent local start --env MY_SECRET_KEY=secret --env OTHER_VAR=$OTHER_VAR
    ```

    Args:
        - agent_config_id (str, optional): An optional agent configuration ID that can be used to set
            configuration based on an agent from a backend API. If set all configuration values will be
            pulled from backend agent configuration.
        - name (str, optional): An optional name to give this agent. Can also be set through
            the environment variable `PREFECT__CLOUD__AGENT__NAME`. Defaults to "agent"
        - labels (List[str], optional): a list of labels, which are arbitrary string
            identifiers used by Prefect Agents when polling for work
        - env_vars (dict, optional): a dictionary of environment variables and values that will
            be set on each flow run that this agent submits for execution
        - max_polls (int, optional): maximum number of times the agent will poll Prefect Cloud
            for flow runs; defaults to infinite
        - agent_address (str, optional):  Address to serve internal api at. Currently this is
            just health checks for use by an orchestration layer. Leave blank for no api server
            (default).
        - no_cloud_logs (bool, optional): Disable logging to a Prefect backend for this agent
            and all deployed flow runs
        - import_paths (List[str], optional): system paths which will be provided to each
            Flow's runtime environment; useful for Flows which import from locally hosted
            scripts or packages
        - show_flow_logs (bool, optional): a boolean specifying whether the agent should
            re-route Flow run logs to stdout; defaults to `False`
        - hostname_label (boolean, optional): a boolean specifying whether this agent should
            auto-label itself with the hostname of the machine it is running on.  Useful for
            flows which are stored on the local filesystem.
    """

    def __init__(
        self,
        agent_config_id: str = None,
        name: str = None,
        labels: Iterable[str] = None,
        env_vars: dict = None,
        import_paths: List[str] = None,
        show_flow_logs: bool = False,
        hostname_label: bool = True,
        max_polls: int = None,
        agent_address: str = None,
        no_cloud_logs: bool = None,
    ) -> None:
        self.processes = set()
        self.import_paths = import_paths or []
        self.show_flow_logs = show_flow_logs
        super().__init__(
            agent_config_id=agent_config_id,
            name=name,
            labels=labels,
            env_vars=env_vars,
            max_polls=max_polls,
            agent_address=agent_address,
            no_cloud_logs=no_cloud_logs,
        )
        hostname = socket.gethostname()

        # Resolve common Docker hostname by using IP
        if hostname == "docker-desktop":
            hostname = socket.gethostbyname(hostname)

        if hostname_label and (hostname not in self.labels):
            assert isinstance(self.labels, list)
            self.labels.append(hostname)

        self.logger.debug(f"Import paths: {self.import_paths}")
        self.logger.debug(f"Show flow logs: {self.show_flow_logs}")

    def heartbeat(self) -> None:
        for process in list(self.processes):
            if process.poll() is not None:
                self.processes.remove(process)
                if process.returncode:
                    self.logger.info(
                        f"Process PID {process.pid} returned non-zero exit code "
                        f"{process.returncode}!"
                    )
        super().heartbeat()

    def deploy_flow(self, flow_run: GraphQLResult) -> str:
        """
        Deploy flow runs on your local machine as subprocesses

        Args:
            - flow_run (GraphQLResult): A GraphQLResult flow run object

        Returns:
            - str: Information about the deployment

        Raises:
            - ValueError: if deployment attempted on unsupported Storage type
        """
        storage = StorageSchema().load(flow_run.flow.storage)
        if isinstance(storage, Docker):
            self.logger.error(
                "Flow run %s has an unsupported storage type: `%s`",
                flow_run.id,
                type(storage).__name__,
            )
            raise TypeError("Unsupported Storage type: %s" % type(storage).__name__)

        run_config = self._get_run_config(flow_run, LocalRun)
        env = self.populate_env_vars(flow_run, run_config=run_config)

        working_dir = None if run_config is None else run_config.working_dir
        if working_dir and not os.path.exists(working_dir):
            msg = f"Flow run {flow_run.id} has a nonexistent `working_dir` configured: {working_dir}"
            self.logger.error(msg)
            raise ValueError(msg)

        stdout = sys.stdout if self.show_flow_logs else DEVNULL

        # note: we will allow these processes to be orphaned if the agent were to exit
        # before the flow runs have completed. The lifecycle of the agent should not
        # dictate the lifecycle of the flow run. However, if the user has elected to
        # show flow logs, these log entries will continue to stream to the users terminal
        # until these child processes exit, even if the agent has already exited.
        p = Popen(
            [sys.executable, "-m", "prefect", "execute", "flow-run"],
            stdout=stdout,
            stderr=STDOUT,
            env=env,
            cwd=working_dir,
        )

        self.processes.add(p)
        self.logger.debug(
            "Submitted flow run {} to process PID {}".format(flow_run.id, p.pid)
        )

        return "PID: {}".format(p.pid)

    def populate_env_vars(
        self, flow_run: GraphQLResult, run_config: LocalRun = None
    ) -> dict:
        """
        Populate metadata and variables in the environment variables for a flow run

        Args:
            - flow_run (GraphQLResult): A flow run object
            - run_config (LocalRun, optional): The `run_config` for the flow, if any.

        Returns:
            - dict: a dictionary representing the populated environment variables
        """
        # Environment variables come from (later options override):

        # 1. Local environment
        env = os.environ.copy()

        # 2. Logging config
        env.update({"PREFECT__LOGGING__LEVEL": config.logging.level})

        # 3. PYTHONPATH
        # XXX: I'm 99% sure this isn't needed, a new Python process should
        # always add the CWD as the first entry in sys.path
        python_path = [
            run_config.working_dir
            if run_config is not None and run_config.working_dir
            else os.getcwd()
        ]
        if os.environ.get("PYTHONPATH"):
            python_path.append(os.environ["PYTHONPATH"])
        if self.import_paths:
            python_path.extend(self.import_paths)
        env["PYTHONPATH"] = os.pathsep.join(python_path)

        # 4. Values set on the agent via `--env`
        env.update(self.env_vars)

        # 5. Values set on a LocalRun RunConfig (if present
        if run_config is not None and run_config.env is not None:
            env.update(run_config.env)

        # 6. Allow logs to be disabled by flow run settings but not enabled
        if not self.log_to_cloud or "PREFECT__CLOUD__SEND_FLOW_RUN_LOGS" not in env:
            env["PREFECT__CLOUD__SEND_FLOW_RUN_LOGS"] = str(self.log_to_cloud).lower()

        # 7. Non-overrideable required env vars
        env.update(
            {
                "PREFECT__BACKEND": config.backend,
                "PREFECT__CLOUD__API": config.cloud.api,
                "PREFECT__CLOUD__API_KEY": self.flow_run_api_key,
                "PREFECT__CLOUD__TENANT_ID": (
                    # Providing a tenant id is only necessary if authenticating
                    self.client.tenant_id
                    if self.flow_run_api_key
                    else None
                ),
                "PREFECT__CLOUD__AGENT__LABELS": str(self.labels),
                "PREFECT__CONTEXT__FLOW_RUN_ID": flow_run.id,  # type: ignore
                "PREFECT__CONTEXT__FLOW_ID": flow_run.flow.id,  # type: ignore
                "PREFECT__CLOUD__USE_LOCAL_SECRETS": "false",
                "PREFECT__ENGINE__FLOW_RUNNER__DEFAULT_CLASS": "prefect.engine.cloud.CloudFlowRunner",
            }
        )

        # Filter out None values
        return {k: v for k, v in env.items() if v is not None}

    @staticmethod
    def generate_supervisor_conf(
        token: str = None,
        labels: Iterable[str] = None,
        env_vars: dict = None,
        import_paths: List[str] = None,
        show_flow_logs: bool = False,
        key: str = None,
        tenant_id: str = None,
        agent_config_id: str = None,
    ) -> str:
        """
        Generate and output an installable supervisorctl configuration file for the agent.

        Args:
            - labels (List[str], optional): a list of labels, which are arbitrary string
                identifiers used by Prefect Agents when polling for work
            - env_vars (dict, optional): a dictionary of environment variables and values that
                will be set on each flow run that this agent submits for execution
            - import_paths (List[str], optional): system paths which will be provided to each
                Flow's runtime environment; useful for Flows which import from locally hosted
                scripts or packages
            - show_flow_logs (bool, optional): a boolean specifying whether the agent should
                re-route Flow run logs to stdout; defaults to `False`
            - key (str, optional): An API key for the agent to use for authentication
                with Prefect Cloud
            - tenant_id (str, optional): A tenant ID for the agent to connect to. If not
                set, the default tenant associated with the API key will be used.
            - agent_config_id (str, optional): An agent config id to link to for health
                check automations

        Returns:
            - str: A string representation of the generated configuration file
        """

        # Use defaults if not provided
        labels = labels or []
        env_vars = env_vars or {}
        import_paths = import_paths or []

        if token and key:
            raise ValueError(
                "Given both a API token and API key. Only one authentication method "
                "may be provided."
            )

        with open(
            os.path.join(os.path.dirname(__file__), "supervisord.conf"), "r"
        ) as conf_file:
            conf = conf_file.read()

        add_opts = ""
        add_opts += "-f " if show_flow_logs else ""
        add_opts += " ".join("-l {label} ".format(label=label) for label in labels)
        add_opts += " ".join(
            "-e {k}={v} ".format(k=k, v=v) for k, v in env_vars.items()
        )
        add_opts += " ".join("-p {path}".format(path=path) for path in import_paths)

        if key:
            add_opts += f"-k {key} "
        if tenant_id:
            add_opts += f"--tenant-id {tenant_id} "
        if agent_config_id:
            add_opts += f"--agent-config-id {agent_config_id}"

        conf = conf.replace("{{OPTS}}", add_opts)
        return conf


if __name__ == "__main__":
    LocalAgent().start()



from typing import Union, Optional

from google.cloud import secretmanager

from prefect.tasks.secrets.base import SecretBase
from prefect.utilities.tasks import defaults_from_attrs


class GCPSecret(SecretBase):
    """
    Task for retrieving a secret from GCP Secrets Manager and returning it as a dictionary.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    For authentication, there are three options: you can set the `GCP_CREDENTIALS` Prefect Secret
    containing your GCP access keys, or [explicitly provide a credentials dictionary],
    or otherwise it will use [default Google client logic].

    [explicitly provide a credentials dictionary]
    https://googleapis.dev/python/google-api-core/latest/auth.html#explicit-credentials

    [default Google client logic]
    https://googleapis.dev/python/google-api-core/latest/auth.html

    Args:
        - project_id (Union[str, int], optional): the name of the project where the Secret is saved
        - secret_id (str, optional): the name of the secret to retrieve
        - version_id (Union[str, int], optional): the version number of the secret to use;
            defaults to 'latest'
        - credentials (dict, optional): dictionary containing GCP credentials
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        project_id: Union[str, int] = None,
        secret_id: str = None,
        version_id: Union[str, int] = "latest",
        credentials: Optional[dict] = None,
        **kwargs,
    ):
        self.project_id = project_id
        self.secret_id = secret_id
        self.version_id = version_id
        self.credentials = credentials
        super().__init__(**kwargs)

    @defaults_from_attrs("project_id", "secret_id", "version_id", "credentials")
    def run(
        self,
        project_id: Union[str, int] = None,
        secret_id: str = None,
        version_id: Union[str, int] = "latest",
        credentials: Optional[dict] = None,
    ) -> str:
        """
        Task run method.

        Args:
            - project_id (Union[str, int], optional): the name of the project where the Secret is saved
            - secret_id (str, optional): the name of the secret to retrieve
            - version_id (Union[str, int], optional): the version number of the secret to use;
                defaults to "latest"
            - credentials (dict, optional): your GCP credentials passed from an upstream Secret task.
                If not provided here default Google client logic will be used.

        Returns:
            - str: the contents of this secret
        """
        if project_id is None:
            raise ValueError("A GCP project ID must be provided.")
        if secret_id is None:
            raise ValueError("A GCP Secret Manager secret ID must be provided.")

        # Create the Secret Manager client.
        client = secretmanager.SecretManagerServiceClient(credentials=credentials)

        # Build the resource name of the secret version.
        name = f"projects/{project_id}/secrets/{secret_id}/versions/{version_id}"

        # Access the secret version.
        response = client.access_secret_version(name=name)

        # Return the decoded payload.
        return response.payload.data.decode("UTF-8")



"""
This module contains a collection of tasks for interacting with Redis via
the redis-py library.
"""

try:
    from prefect.tasks.redis.redis_tasks import RedisSet, RedisGet, RedisExecute
except ImportError as err:
    raise ImportError(
        'Using `prefect.tasks.redis` requires Prefect to be installed with the "redis" extra.'
    ) from err

__all__ = ["RedisExecute", "RedisGet", "RedisSet"]



import time
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional

import azure.identity
import azure.mgmt.datafactory

from prefect import Task
from prefect.client import Secret
from prefect.utilities.tasks import defaults_from_attrs


def _get_datafactory_client(azure_credentials: Dict[str, str]):
    """
    Helper function to create datafactory client.
    """
    client_secret_credential = azure.identity.ClientSecretCredential(
        client_id=azure_credentials["client_id"],
        client_secret=azure_credentials["client_secret"],
        tenant_id=azure_credentials["tenant_id"],
    )
    subscription_id = azure_credentials["subscription_id"]
    datafactory_client = azure.mgmt.datafactory.DataFactoryManagementClient(
        client_secret_credential, subscription_id
    )
    return datafactory_client


class DatafactoryCreate(Task):
    """
    Task for creating an Azure datafactory.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    Args:
        - datafactory_name (str): Name of the datafactory to create.
        - resource_group_name (str): Name of the resource group.
        - azure_credentials_secret (str): Name of the Prefect Secret that stores
            your Azure credentials; This Secret must be JSON string with the keys
            `subscription_id`, `client_id`, `secret`, and `tenant`. Defaults to
            "AZ_CREDENTIALS".
        - location (str, optional): The location of the datafactory.
        - polling_interval (int, optional): The interval, in seconds, to check the provisioning
            state of the datafactory Defaults to 10.
        - options (dict, optional): The options to be passed to the
            `create_or_update` method.
        - **kwargs (dict, optional): Additional keyword arguments to pass to the
            Task constructor.
    """

    def __init__(
        self,
        datafactory_name: str = None,
        resource_group_name: str = None,
        azure_credentials_secret: str = "AZ_CREDENTIALS",
        location: str = "eastus",
        polling_interval: int = 10,
        options: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> None:
        self.datafactory_name = datafactory_name
        self.resource_group_name = resource_group_name
        self.azure_credentials_secret = azure_credentials_secret
        self.location = location
        self.polling_interval = polling_interval
        self.options = options
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "datafactory_name",
        "resource_group_name",
        "azure_credentials_secret",
        "location",
        "polling_interval",
        "options",
    )
    def run(
        self,
        datafactory_name: str = None,
        resource_group_name: str = None,
        azure_credentials_secret: str = None,
        location: Optional[str] = None,
        polling_interval: Optional[int] = None,
        options: Optional[Dict[str, Any]] = None,
    ) -> Dict[Any, Any]:
        """
        Create an Azure datafactory.

        Args:
            - datafactory_name (str): Name of the datafactory to create.
            - resource_group_name (str): Name of the resource group.
            - azure_credentials_secret (str): Name of the Prefect Secret that stores
                your Azure credentials; This Secret must be JSON string with the keys
                `subscription_id`, `client_id`, `secret`, and `tenant`. Defaults to
                "AZ_CREDENTIALS".
            - location (str, optional): The location of the datafactory.
            - polling_interval (int, optional): The interval, in seconds, to check the provisioning
            state of the datafactory Defaults to 10.
            - options (dict, optional): The options to be passed to the
                `create_or_update` method.

        Returns:
            The datafactory name.
        """
        if not datafactory_name:
            raise ValueError("The datafactory_name must be specified.")
        if not resource_group_name:
            raise ValueError("The resource_group_name must be specified.")

        azure_credentials = Secret(azure_credentials_secret).get()
        datafactory_client = _get_datafactory_client(azure_credentials)

        self.logger.info(
            f"Preparing to create the {datafactory_name} datafactory under "
            f"{resource_group_name} in {location}"
        )
        factory = azure.mgmt.datafactory.models.Factory(location=location)
        create_factory = datafactory_client.factories.create_or_update(
            resource_group_name, datafactory_name, factory, **options or {}
        )
        while create_factory.provisioning_state != "Succeeded":
            self.logger.info(
                f"The {datafactory_name} factory status: "
                f"{create_factory.provisioning_state}"
            )
            if create_factory.provisioning_state == "Failed":
                raise RuntimeError(
                    f"Failed to provision the {datafactory_name} factory"
                )
            time.sleep(polling_interval)

        return datafactory_name


class PipelineCreate(Task):
    """
    Task for creating an Azure datafactory pipeline.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    Args:
        - datafactory_name (str): Name of the datafactory to create.
        - resource_group_name (str): Name of the resource group.
        - pipeline_name (str): Name of the pipeline.
        - activities (list): The list of activities to run in the pipeline.
        - azure_credentials_secret (str, optional): Name of the Prefect Secret that stores
            your Azure credentials; This Secret must be JSON string with the keys
            `subscription_id`, `client_id`, `secret`, and `tenant`. Defaults to
            "AZ_CREDENTIALS".
        - parameters (dict): The parameters to be used in pipeline.
        - options (dict, optional): The options to be passed to the
            `create_or_update` method.
        - **kwargs (dict, optional): Additional keyword arguments to pass to the
            Task constructor.
    """

    def __init__(
        self,
        datafactory_name: str = None,
        resource_group_name: str = None,
        pipeline_name: str = None,
        activities: List[Any] = None,
        azure_credentials_secret: str = "AZ_CREDENTIALS",
        parameters: Optional[Dict[str, Any]] = None,
        options: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> None:
        self.datafactory_name = datafactory_name
        self.resource_group_name = resource_group_name
        self.pipeline_name = pipeline_name
        self.activities = activities
        self.azure_credentials_secret = azure_credentials_secret
        self.parameters = parameters
        self.options = options
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "datafactory_name",
        "resource_group_name",
        "pipeline_name",
        "activities",
        "azure_credentials_secret",
        "parameters",
        "options",
    )
    def run(
        self,
        datafactory_name: str = None,
        resource_group_name: str = None,
        pipeline_name: str = None,
        activities: List[Any] = None,
        azure_credentials_secret: str = None,
        parameters: Optional[Dict[str, Any]] = None,
        options: Optional[Dict[str, Any]] = None,
    ) -> Dict[Any, Any]:
        """
        Create an Azure datafactory pipeline.

        Args:
            - datafactory_name (str): Name of the datafactory to create.
            - resource_group_name (str): Name of the resource group.
            - pipeline_name (str): Name of the pipeline.
            - activities (list): The list of activities to run in the pipeline.
            - parameters (dict): The parameters to be used in pipeline.
            - azure_credentials_secret (str, optional): Name of the Prefect Secret that stores
                your Azure credentials; This Secret must be JSON string with the keys
                `subscription_id`, `client_id`, `secret`, and `tenant`. Defaults to
                "AZ_CREDENTIALS".
            - options (dict, optional): The options to be passed to the
                `create_or_update` method.

        Returns:
            The pipeline name.
        """
        if not datafactory_name:
            raise ValueError("The datafactory_name must be specified.")
        if not resource_group_name:
            raise ValueError("The resource_group_name must be specified.")
        if not pipeline_name:
            raise ValueError("The pipeline_name must be specified.")
        if not activities:
            raise ValueError("The activities must be specified.")

        azure_credentials = Secret(azure_credentials_secret).get()
        datafactory_client = _get_datafactory_client(azure_credentials)

        self.logger.info(
            f"Preparing to create the {pipeline_name} pipeline "
            f"containing {len(activities)} activities in the "
            f"{datafactory_name} factory under {resource_group_name}."
        )
        pipeline = azure.mgmt.datafactory.models.PipelineResource(
            activities=activities, parameters=parameters or {}
        )
        datafactory_client.pipelines.create_or_update(
            resource_group_name,
            datafactory_name,
            pipeline_name,
            pipeline,
            **options or {},
        )

        return pipeline_name


class PipelineRun(Task):
    """
    Task for creating an Azure datafactory pipeline run.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    Args:
        - datafactory_name (str): Name of the datafactory to create.
        - resource_group_name (str): Name of the resource group.
        - pipeline_name (str): Name of the pipeline.
        - azure_credentials_secret (str, optional): Name of the Prefect Secret that stores
            your Azure credentials; This Secret must be JSON string with the keys
            `subscription_id`, `client_id`, `secret`, and `tenant`. Defaults to
            "AZ_CREDENTIALS".
        - parameters (dict, optional): The parameters to be used in pipeline.
        - polling_interval (int, optional): The interval, in seconds, to check the status of the run.
            Defaults to 10.
        - last_updated_after (datetime, optional): The time at or after which the run event
            was updated; used to filter and query the pipeline run, and defaults to yesterday.
        - last_updated_before (datetime, optional): The time at or before which the run event
            was updated; used to filter and query the pipeline run and defaults to tomorrow.
        - **kwargs (dict, optional): Additional keyword arguments to pass to the
            Task constructor.
    """

    def __init__(
        self,
        datafactory_name: str = None,
        resource_group_name: str = None,
        pipeline_name: str = None,
        azure_credentials_secret: str = "AZ_CREDENTIALS",
        parameters: Optional[Dict[str, Any]] = None,
        polling_interval: int = 10,
        last_updated_after: Optional[datetime] = None,
        last_updated_before: Optional[datetime] = None,
        **kwargs,
    ) -> None:
        self.datafactory_name = datafactory_name
        self.resource_group_name = resource_group_name
        self.pipeline_name = pipeline_name
        self.azure_credentials_secret = azure_credentials_secret
        self.parameters = parameters
        self.polling_interval = polling_interval
        self.last_updated_after = last_updated_after
        self.last_updated_before = last_updated_before
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "datafactory_name",
        "resource_group_name",
        "pipeline_name",
        "azure_credentials_secret",
        "parameters",
        "polling_interval",
        "last_updated_after",
        "last_updated_before",
    )
    def run(
        self,
        datafactory_name: str = None,
        resource_group_name: str = None,
        pipeline_name: str = None,
        azure_credentials_secret: str = None,
        parameters: Optional[Dict[str, Any]] = None,
        polling_interval: Optional[int] = None,
        last_updated_after: Optional[datetime] = None,
        last_updated_before: Optional[datetime] = None,
    ) -> Dict[Any, Any]:
        """
        Create an Azure datafactory pipeline run.

        Args:
            - datafactory_name (str): Name of the datafactory to create.
            - resource_group_name (str): Name of the resource group.
            - pipeline_name (str): Name of the pipeline.
            - azure_credentials_secret (str, optional): Name of the Prefect Secret that stores
                your Azure credentials; This Secret must be JSON string with the keys
                `subscription_id`, `client_id`, `secret`, and `tenant`. Defaults to
                "AZ_CREDENTIALS".
            - parameters (dict, optional): The parameters to be used in pipeline.
            - polling_interval (int, optional): The interval, in seconds, to check the status of the run.
                Defaults to 10.
            - last_updated_after (datetime, optional): The time at or after which the run event
                was updated; used to filter and query the pipeline run, and defaults to yesterday.
            - last_updated_before (datetime, optional): The time at or before which the run event
                was updated; used to filter and query the pipeline run and defaults to tomorrow.

        Returns:
            The pipeline run response.
        """
        if not datafactory_name:
            raise ValueError("The datafactory_name must be specified.")
        if not resource_group_name:
            raise ValueError("The resource_group_name must be specified.")
        if not pipeline_name:
            raise ValueError("The pipeline_name must be specified.")

        azure_credentials = Secret(azure_credentials_secret).get()
        datafactory_client = _get_datafactory_client(azure_credentials)
        last_updated_after = last_updated_after or datetime.utcnow() - timedelta(days=1)
        last_updated_before = last_updated_before or datetime.utcnow() + timedelta(
            days=1
        )

        self.logger.info(
            f"Preparing to run the {pipeline_name} pipeline in the "
            f"{datafactory_name} factory under {resource_group_name}."
        )
        # Create a pipeline run
        create_run = datafactory_client.pipelines.create_run(
            resource_group_name,
            datafactory_name,
            pipeline_name,
            parameters=parameters or {},
        )

        # Monitor the pipeline run
        pipeline_run = datafactory_client.pipeline_runs.get(
            resource_group_name, datafactory_name, create_run.run_id
        )
        while pipeline_run.status != "Succeeded":
            pipeline_run = datafactory_client.pipeline_runs.get(
                resource_group_name, datafactory_name, create_run.run_id
            )
            self.logger.info(f"The pipeline run status: {pipeline_run.status}")
            if pipeline_run.status.lower() not in ["queued", "inprogress", "succeeded"]:
                raise RuntimeError(pipeline_run.message)
            time.sleep(polling_interval)

        filter_params = azure.mgmt.datafactory.models.RunFilterParameters(
            last_updated_after=last_updated_after,
            last_updated_before=last_updated_before,
        )
        query_response = datafactory_client.activity_runs.query_by_pipeline_run(
            resource_group_name, datafactory_name, pipeline_run.run_id, filter_params
        )
        return query_response



from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Extra, Field, validator


class TaskDependency(BaseModel):
    task_key: str


class AutoScale(BaseModel):
    min_workers: int
    max_workers: int


class AwsAvailability(Enum):
    SPOT = "SPOT"
    ON_DEMAND = "ON_DEMAND"
    SPOT_WITH_FALLBACK = "SPOT_WITH_FALLBACK"


class EbsVolumeType(Enum):
    GENERAL_PURPOSE_SSD = "GENERAL_PURPOSE_SSD"
    THROUGHPUT_OPTIMIZED_HDD = "THROUGHPUT_OPTIMIZED_HDD"


class AwsAttributes(BaseModel):
    first_on_demand: Optional[int] = None
    availability: Optional[AwsAvailability] = None
    zone_id: Optional[str] = None
    instance_profile_arn: Optional[str] = None
    spot_bid_price_percent: Optional[int] = None
    ebs_volume_type: Optional[EbsVolumeType] = None
    ebs_volume_count: Optional[int] = None
    ebs_volume_size: Optional[int] = None
    ebs_volume_iops: Optional[int] = None
    ebs_volume_throughput: Optional[int] = None


class DbfsStorageInfo(BaseModel):
    destination: Optional[str] = None


class S3StorageInfo(BaseModel):
    destination: Optional[str] = None
    region: Optional[str] = None
    endpoint: Optional[str] = None
    enable_encryption: Optional[bool] = None
    encryption_type: Optional[str] = None
    kms_key: Optional[str] = None
    canned_acl: Optional[str] = None


class FileStorageInfo(BaseModel):
    destination: Optional[str] = None


class ClusterLogConf(BaseModel):
    dbfs: Optional[DbfsStorageInfo] = None
    s3: Optional[S3StorageInfo] = None


class InitScriptInfo(BaseModel):
    dbfs: Optional[DbfsStorageInfo] = None
    file: Optional[FileStorageInfo] = None
    S3: Optional[S3StorageInfo] = None


class NewCluster(BaseModel, extra=Extra.allow):
    spark_version: str
    spark_conf: Dict = Field(default_factory=dict)
    node_type_id: Optional[str] = None
    autoscale: Optional[AutoScale] = None
    num_workers: Optional[int] = None
    aws_attributes: Optional[AwsAttributes] = None
    driver_node_type_id: Optional[str] = None
    ssh_public_keys: Optional[List[str]] = None
    custom_tags: Optional[Dict[str, Any]] = None
    cluster_log_conf: Optional[ClusterLogConf] = None
    init_scripts: Optional[List[InitScriptInfo]] = None
    spark_env_vars: Optional[Dict] = None
    enable_elastic_disk: Optional[bool] = None
    driver_instance_pool_id: Optional[str] = None
    instance_pool_id: Optional[str] = None
    policy_id: Optional[str] = None
    data_security_mode: Optional[str] = None
    single_user_name: Optional[str] = None

    @validator("instance_pool_id", always=True, pre=True)
    def node_type_or_instance_pool(cls, instance_pool_id, values):
        node_type_id = values.get("node_type_id")
        if instance_pool_id is not None and node_type_id is not None:
            raise ValueError(
                "Cannot specify instance_pool_id if node_type_id has been specified"
            )

        if instance_pool_id is None and node_type_id is None:
            raise ValueError("Must specify either instance_pool_id or node_type_id")
        return instance_pool_id


class NotebookTask(BaseModel):
    notebook_path: str
    base_parameters: Optional[Dict[str, Any]] = None


class SparkJarTask(BaseModel):
    main_class_name: str
    parameters: List[str] = Field(default_factory=list)


class SparkPythonTask(BaseModel):
    python_file: str
    parameters: List[str] = Field(default_factory=list)


class SparkSubmitTask(BaseModel):
    parameters: List[str]


class PipelineTask(BaseModel):
    pipeline_id: str


class PythonWheelTask(BaseModel):
    package_name: str
    entry_point: Optional[str] = None
    parameters: List[str] = Field(default_factory=list)


class MavenLibrary(BaseModel):
    coordinates: str
    repo: Optional[str] = None
    exclusions: List[str] = Field(default_factory=list)


class PythonPyPiLibrary(BaseModel):
    package: str
    repo: Optional[str] = None


class RCranLibrary(BaseModel):
    package: str
    repo: Optional[str] = None


class Library(BaseModel):
    jar: Optional[str] = None
    egg: Optional[str] = None
    whl: Optional[str] = None
    pypi: Optional[PythonPyPiLibrary] = None
    maven: Optional[MavenLibrary] = None
    cran: Optional[RCranLibrary] = None


class JobEmailNotifications(BaseModel):
    on_start: List[str] = Field(default_factory=list)
    on_success: List[str] = Field(default_factory=list)
    on_failure: List[str] = Field(default_factory=list)
    no_alert_for_skipped_runs: bool = False


class JobTaskSettings(BaseModel):
    task_key: str
    description: Optional[str] = None
    depends_on: List[TaskDependency] = Field(default_factory=list)
    existing_cluster_id: Optional[str] = None
    new_cluster: Optional[NewCluster] = None
    job_cluster_key: Optional[str] = None
    notebook_task: Optional[NotebookTask] = None
    spark_jar_task: Optional[SparkJarTask] = None
    spark_python_task: Optional[SparkPythonTask] = None
    spark_submit_task: Optional[SparkSubmitTask] = None
    pipeline_task: Optional[PipelineTask] = None
    python_wheel_task: Optional[PythonWheelTask] = None
    libraries: Optional[List[Library]] = None
    email_notifications: Optional[JobEmailNotifications] = None
    timeout_seconds: Optional[int] = None
    max_retries: Optional[int] = None
    min_retry_interval_millis: Optional[int] = None
    retry_on_timeout: Optional[bool] = None


class JobCluster(BaseModel):
    job_cluster_key: str
    new_cluster: Optional[NewCluster] = None


class CanManage(Enum):
    CAN_MANAGE = "CAN_MANAGE"


class CanManageRun(Enum):
    CAN_MANAGE_RUN = "CAN_MANAGE_RUN"


class CanView(Enum):
    CAN_VIEW = "CAN_VIEW"


class IsOwner(Enum):
    IS_OWNER = "IS_OWNER"


class PermissionLevel(BaseModel):
    __root__: Union[CanManage, CanManageRun, CanView, IsOwner]


class PermissionLevelForGroup(BaseModel):
    __root__: Union[CanManage, CanManageRun, CanView]


class AccessControlRequestForUser(BaseModel):
    class Config:
        use_enum_values = True

    user_name: str
    permission_level: Union[CanManage, CanManageRun, CanView, IsOwner]


class AccessControlRequestForGroup(BaseModel):
    class Config:
        use_enum_values = True

    group_name: str
    permission_level: Union[CanManage, CanManageRun, CanView]


class AccessControlRequest(BaseModel):
    __root__: Union[AccessControlRequestForUser, AccessControlRequestForGroup]


class GitSource(BaseModel, extra=Extra.allow):
    git_url: str
    git_provider: str
    git_branch: Optional[str] = None
    git_tag: Optional[str] = None
    git_commit: Optional[str] = None

    @validator("git_tag", always=True)
    def branch_or_tag(cls, v, values):
        if "git_branch" in values is not None and v:
            raise ValueError("Cannot specify git tag if git branch has been specified")
        return v

    @validator("git_branch", always=True)
    def commit_or_branch(cls, v, values):
        if "git_commit" in values is not None and v:
            raise ValueError(
                "Cannot specify git branch if git commit has been specified"
            )
        return v

    @validator("git_commit", always=True)
    def tag_or_commit(cls, v, values):
        if "git_tag" in values is not None and v:
            raise ValueError("Cannot specify git commit if git tag has been specified")
        return v



__all__ = [
    "create_project",
    "create_exam_pool",
    "create_pool",
    "create_tasks",
    "open_pool",
    "open_exam_pool",
    "wait_pool",
    "get_assignments",
    "get_assignments_df",
    "accept_assignment",
    "reject_assignment",
]

import logging
import http
import pandas as pd
import time
from datetime import datetime, timedelta
from enum import Enum, unique
from typing import Dict, List, Optional, Union

from prefect import task

from toloka.client import Assignment, Pool, Project, TolokaClient, Training
from toloka.client import Task as TolokaTask
from toloka.client.analytics_request import CompletionPercentagePoolAnalytics
from toloka.client.assignment import GetAssignmentsTsvParameters
from toloka.client.batch_create_results import TaskBatchCreateResult
from toloka.client.exceptions import IncorrectActionsApiError

from .utils import extract_id, structure_from_conf, with_logger, with_toloka_client


@task
@with_toloka_client
def create_project(
    obj: Union[Project, Dict, str, bytes],
    *,
    toloka_client: Optional[TolokaClient] = None,
) -> Project:
    """
    Task to create a Toloka `Project` object from given config.

    Args:
        - obj (Project, Dict, str, bytes): Either a `Project` object itself
            or a config to make a `Project`.
        - token (str, optional): Toloka token value provided with a Prefect Secret.
            If not provided the task will fallback to using Secret with secret_name.
        - secret_name (str, optional): Allow to use non-default secret for Toloka token.
            Default: "TOLOKA_TOKEN".
        - env (str, optional): Allow to use non-default Toloka environment.
            Default: "PRODUCTION".

    Returns:
        - Project: Toloka project object with id assigned.

    Example:
        >>> project_conf = {...}  # May also be configured toloka.client.Project object.
        >>> project = create_project(project_conf)
        ...
    """
    obj = structure_from_conf(obj, Project)
    return toloka_client.create_project(obj)


@task
@with_toloka_client
def create_exam_pool(
    obj: Union[Training, Dict, str, bytes],
    *,
    project_id: Union[Project, Dict, str, None] = None,
    toloka_client: Optional[TolokaClient] = None,
) -> Training:
    """
    Task to create a Toloka `Training` pool object from given config.

    Args:
        - obj (Training, Dict, str, bytes): Either a `Training` object itself
            or a config to make a `Training`.
        - project_id (Project, Dict, str, optional): Project ID to assign a training pool to.
            May pass either an object, config or project_id value.
        - token (str, optional): Toloka token value provided with a Prefect Secret.
            If not provided the task will fallback to using Secret with secret_name.
        - secret_name (str, optional): Allow to use non-default secret for Toloka token.
            Default: "TOLOKA_TOKEN".
        - env (str, optional): Allow to use non-default Toloka environment.
            Default: "PRODUCTION".

    Returns:
        - Training: Toloka training pool object with id assigned.

    Example:
        >>> project = create_project({...})
        >>> exam = create_exam_pool({...}, project_id=project)
        ...
    """
    obj = structure_from_conf(obj, Training)
    if project_id is not None:
        obj.project_id = extract_id(project_id, Project)
    return toloka_client.create_training(obj)


@task
@with_toloka_client
def create_pool(
    obj: Union[Pool, Dict, str, bytes],
    *,
    project_id: Union[Project, Dict, str, None] = None,
    exam_pool_id: Union[Training, Dict, str, None] = None,
    expiration: Union[datetime, timedelta, None] = None,
    reward_per_assignment: Optional[float] = None,
    toloka_client: Optional[TolokaClient] = None,
) -> Pool:
    """
    Task to create a Toloka `Pool` object from given config.

    Args:
        - obj (Pool, Dict, str, bytes): Either a `Pool` object itself or a config to make a `Pool`.
        - project_id (Project, Dict, str, optional): Project ID to assign a pool to.
            May pass either an object, config or project_id value.
        - exam_pool_id (Training, Dict, str, optional): Related training pool ID.
            May pass either an object, config or pool_id value.
        - expiration (datetime, timedelta, optional): Expiration setting. May pass any of:
            `None` if this setting if already present;
            `datetime` object to set exact datetime;
            `timedelta` to set expiration related to the current time.
        - reward_per_assignment (float, optional): Allow to redefine reward per assignment.
        - token (str, optional): Toloka token value provided with a Prefect Secret.
            If not provided the task will fallback to using Secret with secret_name.
        - secret_name (str, optional): Allow to use non-default secret for Toloka token.
            Default: "TOLOKA_TOKEN".
        - env (str, optional): Allow to use non-default Toloka environment.
            Default: "PRODUCTION".

    Returns:
        - Pool: Toloka pool object with id assigned.

    Example:
        >>> project = create_project({...})
        >>> exam = create_exam_pool({...}, project_id=project)
        >>> pool = create_pool({...}, project_id=project, exam_pool_id=exam)
        ...
    """
    obj = structure_from_conf(obj, Pool)
    if project_id is not None:
        obj.project_id = extract_id(project_id, Project)
    if exam_pool_id:
        if obj.quality_control.training_requirement is None:
            raise ValueError(
                "pool.quality_control.training_requirement should be set before exam_pool assignment"
            )
        obj.quality_control.training_requirement.training_pool_id = extract_id(
            exam_pool_id, Training
        )
    if expiration:
        if isinstance(expiration, timedelta):
            obj.will_expire = datetime.utcnow() + expiration
        else:
            obj.will_expire = expiration
    if reward_per_assignment:
        obj.reward_per_assignment = reward_per_assignment
    return toloka_client.create_pool(obj)


@task
@with_toloka_client
def create_tasks(
    tasks: List[Union[TolokaTask, Dict]],
    *,
    pool_id: Union[Pool, Training, Dict, str, None] = None,
    allow_defaults: bool = False,
    open_pool: bool = False,
    skip_invalid_items: bool = False,
    toloka_client: Optional[TolokaClient] = None,
) -> TaskBatchCreateResult:
    """
    Task to create a list of tasks for a given pool.

    Args:
        - tasks (List[Union[TolokaTask, Dict]]): List of either a `toloka.client.Task` objects
            or task conofigurations.
        - pool_id (Pool, Training, Dict, str, optional): Allow to set tasks pool ID
            if it's not present in tasks themselves.
            May be either a `Pool` or `Training` object or config or a pool_id value.
        - allow_defaults (bool, optional): Allow to use the overlap that is set in the pool parameters.
        - open_pool (bool, optional): Open the pool immediately after creating a task suite, if closed.
        - skip_invalid_items (bool, optional): Allow to skip invalid tasks.
            You can handle them using resulting TaskBatchCreateResult object.
        - token (str, optional): Toloka token value provided with a Prefect Secret.
            If not provided the task will fallback to using Secret with secret_name.
        - secret_name (str, optional): Allow to use non-default secret for Toloka token.
            Default: "TOLOKA_TOKEN".
        - env (str, optional): Allow to use non-default Toloka environment.
            Default: "PRODUCTION".

    Returns:
        - TaskBatchCreateResult: Result object.

    Example:
        >>> tasks = [{'input_values': ...}, {'input_values': ...}, {'input_values': ...}]
        >>> create_tasks(tasks,
        ...              pool_id='some-pool_id-123',
        ...              open_pool=True,
        ...              allow_defaults=True)
        ...
    """
    tasks = [structure_from_conf(task, TolokaTask) for task in tasks]
    if pool_id is not None:
        try:
            pool_id = extract_id(pool_id, Pool)
        except Exception:
            pool_id = extract_id(pool_id, Training)
        for task_ in tasks:
            task_.pool_id = pool_id
    kwargs = {
        "allow_defaults": allow_defaults,
        "open_pool": open_pool,
        "skip_invalid_items": skip_invalid_items,
    }
    return toloka_client.create_tasks(tasks, **kwargs)


@task
@with_toloka_client
def open_pool(
    obj: Union[Pool, Dict, str],
    *,
    toloka_client: Optional[TolokaClient] = None,
) -> Pool:
    """
    Task to open given Toloka pool.

    Args:
        - obj (Pool, Dict, str): Pool id or `Pool` object of it's config.
        - token (str, optional): Toloka token value provided with a Prefect Secret.
            If not provided the task will fallback to using Secret with secret_name.
        - secret_name (str, optional): Allow to use non-default secret for Toloka token.
            Default: "TOLOKA_TOKEN".
        - env (str, optional): Allow to use non-default Toloka environment.
            Default: "PRODUCTION".

    Returns:
        - Pool: Opened pool object.

    Example:
        >>> pool = create_pool({...})
        >>> _tasks_creation = create_tasks([...], pool=pool)
        >>> open_pool(pool, upstream_tasks=[_tasks_creation])
        ...
    """
    pool_id = extract_id(obj, Pool)
    return toloka_client.open_pool(pool_id)


@task
@with_toloka_client
def open_exam_pool(
    obj: Union[Training, str],
    *,
    toloka_client: Optional[TolokaClient] = None,
) -> Pool:
    """
    Task to open given training pool.

    Args:
        - obj (Training, Dict, str): Training pool_id or `Training` object of it's config.
        - token (str, optional): Toloka token value provided with a Prefect Secret.
            If not provided the task will fallback to using Secret with secret_name.
        - secret_name (str, optional): Allow to use non-default secret for Toloka token.
            Default: "TOLOKA_TOKEN".
        - env (str, optional): Allow to use non-default Toloka environment.
            Default: "PRODUCTION".

    Returns:
        - Training: Opened training (exam) pool object.

    Example:
        >>> exam = create_training({...})
        >>> _exam_tasks_creation = create_tasks([...], pool=exam)
        >>> open_exam_pool(exam, upstream_tasks=[_exam_tasks_creation])
        ...
    """
    pool_id = extract_id(obj, Training)
    return toloka_client.open_training(pool_id)


@task
@with_toloka_client
@with_logger
def wait_pool(
    pool_id: Union[Pool, Dict, str],
    period: timedelta = timedelta(seconds=60),
    *,
    open_pool: bool = False,
    logger: logging.Logger,
    toloka_client: Optional[TolokaClient] = None,
) -> Pool:
    """
    Task to wait given Toloka pool until close.

    Args:
        - pool_id (Pool, Dict, str): Either a `Pool` object or it's config or a pool ID value.
        - period (timedelta): Interval between checks. One minute by default.
        - open_pool (bool, optional): Allow to open pool at start if it's closed. False by default.
        - token (str, optional): Toloka token value provided with a Prefect Secret.
            If not provided the task will fallback to using Secret with secret_name.
        - secret_name (str, optional): Allow to use non-default secret for Toloka token.
            Default: "TOLOKA_TOKEN".
        - env (str, optional): Allow to use non-default Toloka environment.
            Default: "PRODUCTION".

    Returns:
        - Pool: Toloka pool object.

    Example:
        >>> pool = create_pool({...})
        >>> _tasks_creation = create_tasks([...], pool=pool)
        >>> wait_pool(pool, open_pool=True, upstream_tasks=[_tasks_creation])
        ...
    """
    pool_id = extract_id(pool_id, Pool)
    pool = toloka_client.get_pool(pool_id)
    if pool.is_closed() and open_pool:
        pool = toloka_client.open_pool(pool_id)

    while pool.is_open():
        op = toloka_client.get_analytics(
            [CompletionPercentagePoolAnalytics(subject_id=pool_id)]
        )
        percentage = toloka_client.wait_operation(op).details["value"][0]["result"][
            "value"
        ]
        logger.info(f"Pool {pool_id} - {percentage}%")

        time.sleep(period.total_seconds())
        pool = toloka_client.get_pool(pool_id)

    return pool


@task
@with_toloka_client
def get_assignments(
    pool_id: Union[Pool, Dict, str],
    status: Union[
        str, List[str], Assignment.Status, List[Assignment.Status], None
    ] = None,
    *,
    toloka_client: Optional[TolokaClient] = None,
    **kwargs,
) -> List[Assignment]:
    """
    Task to get all assignments of selected status from Toloka pool.

    Args:
        - pool_id (Pool, Dict, str): Either a `Pool` object or it's config or a pool ID value.
        - status (str, List[str], optional): A status or a list of statuses to get.
            All statuses (None) by default.
        - token (str, optional): Toloka token value provided with a Prefect Secret.
            If not provided the task will fallback to using Secret with secret_name.
        - secret_name (str, optional): Allow to use non-default secret for Toloka token.
            Default: "TOLOKA_TOKEN".
        - env (str, optional): Allow to use non-default Toloka environment.
            Default: "PRODUCTION".
        - **kwargs: Other args presented in `toloka.client.search_requests.AssignmentSearchRequest`.

    Returns:
        - List[Assignment]: Get assignments result.

    Example:
        >>> pool = create_pool({...})
        >>> _tasks_creation = create_tasks([...], pool=pool)
        >>> _waiting = wait_pool(pool, open_pool=True, upstream_tasks=[_tasks_creation])
        >>> results = get_assignments(pool, status=['SUBMITTED'], upstream_tasks=[_waiting])
        ...
    """
    pool_id = extract_id(pool_id, Pool)
    return list(toloka_client.get_assignments(pool_id=pool_id, status=status, **kwargs))


@task
@with_toloka_client
def get_assignments_df(
    pool_id: Union[Pool, Dict, str],
    status: Union[
        str, List[str], Assignment.Status, List[Assignment.Status], None
    ] = None,
    *,
    toloka_client: Optional[TolokaClient] = None,
    start_time_from: Optional[datetime] = None,
    start_time_to: Optional[datetime] = None,
    exclude_banned: bool = False,
    field: Optional[List[GetAssignmentsTsvParameters.Field]] = None,
) -> pd.DataFrame:
    """
    Task to get pool assignments of selected statuses
    in Pandas `DataFrame` format useful for aggregation.

    Args:
        - pool_id (Pool, Dict, str): Either a `Pool` object or it's config or a pool ID value.
        - status (str, List[str], optional): A status or a list of statuses to get.
            All statuses (None) by default.
        - token (str, optional): Toloka token value provided with a Prefect Secret.
            If not provided the task will fallback to using Secret with secret_name.
        - secret_name (str, optional): Allow to use non-default secret for Toloka token.
            Default: "TOLOKA_TOKEN".
        - env (str, optional): Allow to use non-default Toloka environment.
            Default: "PRODUCTION".
        - start_time_from (str, optional): Upload assignments submitted after the specified datetime.
        - start_time_to (str, optional): Upload assignments submitted before the specified datetime.
        - exclude_banned (bool, optional): Exclude answers from banned performers,
            even if assignments in suitable status "ACCEPTED".
        - field (List[GetAssignmentsTsvParameters.Field], optional): Select some additional fields.
            You can find possible values in the
            `toloka.client.assignment.GetAssignmentsTsvParameters.Field` enum.

    Returns:
        - DataFrame: `pd.DataFrame` with selected assignments.
            Note that nested paths are presented with a ":" sign.

    Example:
        >>> pool = create_pool({...})
        >>> _tasks_creation = create_tasks([...], pool=pool)
        >>> _waiting = wait_pool(pool, open_pool=True, upstream_tasks=[_tasks_creation])
        >>> df = get_assignments_df(pool, status=['SUBMITTED'], upstream_tasks=[_waiting])
        ...
    """
    pool_id = extract_id(pool_id, Pool)
    if not status:
        status = []
    elif isinstance(status, (str, Assignment.Status)):
        status = [status]
    kwargs = {
        "start_time_from": start_time_from,
        "start_time_to": start_time_to,
        "exclude_banned": exclude_banned,
        "field": field,
    }
    return toloka_client.get_assignments_df(
        pool_id=pool_id,
        status=status,
        **{key: value for key, value in kwargs.items() if value is not None},
    )


@unique
class _PatchAssignmentMethod(Enum):
    ACCEPT_ASSIGNMENT = "accept_assignment"
    REJECT_ASSIGNMENT = "reject_assignment"


def _patch_assignment(
    method_name: _PatchAssignmentMethod,
    assignment_id: Union[Assignment, Dict, str],
    public_comment: str,
    fail_if_already_set: bool,
    logger: logging.Logger,
    toloka_client: Optional[TolokaClient],
) -> Union[Assignment, Dict, str]:
    """Function to either accept or reject assignment with given comment."""
    assignment_id = extract_id(assignment_id, Assignment)
    method = getattr(toloka_client, method_name.value)
    try:
        method(assignment_id, public_comment=public_comment)
    except IncorrectActionsApiError as exc:
        logger.warning("Can't %s %s: %s", method_name.value, assignment_id, exc)
        if fail_if_already_set or exc.status_code != http.client.CONFLICT.value:
            raise
    return assignment_id


@task
@with_toloka_client
@with_logger
def accept_assignment(
    assignment_id: Union[Assignment, Dict, str],
    public_comment: str,
    fail_if_already_set: bool = False,
    *,
    logger: logging.Logger,
    toloka_client: Optional[TolokaClient] = None,
) -> Union[Assignment, Dict, str]:
    """
    Task to accept given assignment by given ID.
    Use `accept_assignment.map` to process multiple assignments.
    Pass public_comment with `unmapped` wrapper in this case.

    Args:
        - assignment_id (Assignment, Dict, str): Either an `Assignment` object
            or it's config or an assignment ID value.
        - public_comment (str): Public comment.
        - fail_if_already_set (bool, optional): Fail if the assignment has already been accepted
            (in the UI for example). Skip by default.
        - token (str, optional): Toloka token value provided with a Prefect Secret.
            If not provided the task will fallback to using Secret with secret_name.
        - secret_name (str, optional): Allow to use non-default secret for Toloka token.
            Default: "TOLOKA_TOKEN".
        - env (str, optional): Allow to use non-default Toloka environment.
            Default: "PRODUCTION".

    Returns:
        - Union[Assignment, Dict, str]: The same assignment_id, that was given.

    Example:
        >>> aggregated = do_some_aggregation(assignments)
        >>> to_accept = some_filter_to_accept(aggregated)
        >>> accept_assignment.map(to_accept, unmapped('Well done!'))
        ...
    """
    return _patch_assignment(
        _PatchAssignmentMethod.ACCEPT_ASSIGNMENT,
        assignment_id,
        public_comment,
        fail_if_already_set,
        logger,
        toloka_client,
    )


@task
@with_toloka_client
@with_logger
def reject_assignment(
    assignment_id: Union[Assignment, Dict, str],
    public_comment: str,
    fail_if_already_set: bool = False,
    *,
    logger: logging.Logger,
    toloka_client: Optional[TolokaClient] = None,
) -> Union[Assignment, Dict, str]:
    """
    Task to reject given assignment by given ID.
    Use `reject_assignment.map` to process multiple assignments.
    Pass public_comment with `unmapped` wrapper in this case.

    Args:
        - assignment_id (Assignment, Dict, str): Either an `Assignment` object
            or it's config or an assignment ID value.
        - public_comment (str): Public comment.
        - fail_if_already_set (bool, optional): Fail if the assignment has already been rejected
            (in the UI for example). Skip by default.
        - token (str, optional): Toloka token value provided with a Prefect Secret.
            If not provided the task will fallback to using Secret with secret_name.
        - secret_name (str, optional): Allow to use non-default secret for Toloka token.
            Default: "TOLOKA_TOKEN".
        - env (str, optional): Allow to use non-default Toloka environment.
            Default: "PRODUCTION".

    Returns:
        - Union[Assignment, Dict, str]: The same assignment_id, that was given.

    Example:
        >>> aggregated = do_some_aggregation(assignments)
        >>> to_reject = some_filter_to_reject(aggregated)
        >>> reject_assignment.map(to_reject, unmapped('Incorrect answer'))
        ...
    """
    return _patch_assignment(
        _PatchAssignmentMethod.REJECT_ASSIGNMENT,
        assignment_id,
        public_comment,
        fail_if_already_set,
        logger,
        toloka_client,
    )



class PrefectSignal(BaseException):
    """
    Signals inherit from `BaseException` and will not be caught by normal error
    handling. This allows us to bypass typical error handling by raising signals.

    See `prefect.engine.signals` for additional subclasses used for raising state
    transitions.

    Args:
        - message: A message with additional information about the error
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message)


class VersionLockMismatchSignal(PrefectSignal):
    """
    Raised when version locking is enabled and a task run state version sent to Cloud
    does not match the version expected by the server.

    This is not backwards compatible with `prefect.utilities.exceptions.VersionLockError`

    Args:
        - message: A message with additional information about the error
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message)


class TaskTimeoutSignal(PrefectSignal):
    """
    Raised when a task reaches a timeout limit

    This is not backwards compatible with `prefect.utilities.exceptions.TaskTimeoutError`

    Args:
        - message: A message with additional information about the error
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message)


class PrefectException(Exception):
    """
    The base exception type for all Prefect related exceptions

    Args:
        - message: A message with additional information about the error
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message)


class ClientError(PrefectException):
    """
    Raised when there is error in Prefect Client <-> Server communication

    Args:
        - message: A message with additional information about the error
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message)


class ObjectNotFoundError(ClientError):
    """
    Raised when an object is not found on the Server

    Args:
        - message: A message with additional information about the error
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message)


class AuthorizationError(ClientError):
    """
    Raised when there is an issue authorizing with Prefect Cloud

    Args:
        - message: A message with additional information about the error
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message)


class FlowStorageError(PrefectException):
    """
    Raised when there is an error loading a flow from storage

    Args:
        - message: A message with additional information about the error
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message)



"""
Great Expectations checkpoints are combinations of data source, expectation suite, and
validation operators configuration that can be used to run Great Expectations actions.
Checkpoints are the preferred deployment of validation configuration; you can read more about
setting up checkpoints [at the Great Expectation
docs](https://docs.greatexpectations.io/en/latest/tutorials/getting_started/set_up_your_first_checkpoint.html#set-up-your-first-checkpoint).

You can use these task library tasks to interact with your Great Expectations checkpoint from a
Prefect flow.
"""
from typing import Optional

import great_expectations as ge
from great_expectations.core import RunIdentifier
from great_expectations.checkpoint import Checkpoint
from packaging import version

import prefect
from prefect import Task
from prefect.backend.artifacts import create_markdown_artifact
from prefect.engine import signals
from prefect.utilities.tasks import defaults_from_attrs


class RunGreatExpectationsValidation(Task):
    """
    Task for running data validation with Great Expectations.
    Works with both the Great Expectations v2 (batch_kwargs) and v3 (Batch Request) APIs.

    Example using the GE getting started tutorial:
    https://github.com/superconductive/ge_tutorials/tree/main/getting_started_tutorial_final_v3_api

    The task can be used to run validation in one of the following ways:

    1. checkpoint_name: the name of a pre-configured checkpoint (which bundles expectation suites
    and batch_kwargs). This is the preferred option.
    2. expectation_suite AND batch_kwargs, where batch_kwargs is a dict. This will only work with the
    Great Expectations v2 API.
    3. assets_to_validate: a list of dicts of expectation_suite + batch_kwargs. This will only work
    with the Great Expectations v2 API.

    To create a checkpoint you can use:
    - for the v2 API: `great_expectations checkpoint new <expectations_suite_name> <checkpoint_name>`
    - for the v3 API: `great_expectations --v3-api checkpoint new <checkpoint_name>`

    Here is an example that can be used with both v2 and v3 API provided that
    the checkpoint has been already created, as described above:
    ```python
    from prefect import Flow, Parameter
    from prefect.tasks.great_expectations import RunGreatExpectationsValidation

    validation_task = RunGreatExpectationsValidation()

    with Flow("ge_test") as flow:
        checkpoint_name = Parameter("checkpoint_name")
        prev_run_row_count = 100  # can be taken eg. from Prefect KV store
        validation_task(
            checkpoint_name=checkpoint_name,
            evaluation_parameters=dict(prev_run_row_count=prev_run_row_count),
        )

    flow.run(parameters={"checkpoint_name": "my_checkpoint"})
    ```


    Args:
        - checkpoint_name (str, optional): the name of a pre-configured checkpoint; should match the
            filename of the checkpoint without the extension. Either checkpoint_name or checkpoint
            is required when using the Great Expectations v3 API.
        - ge_checkpoint (Checkpoint, optional): an in-memory GE `Checkpoint` object used to perform
            validation. If not provided then `checkpoint_name` will be used to load the specified
            checkpoint. Either checkpoint_name or checkpoint is required when using the Great
            Expectations v3 API.
        - checkpoint_kwargs (Dict, optional): A dictionary whose keys match the parameters of
            `CheckpointConfig` which can be used to update and populate the task's Checkpoint at runtime.
            Only used in the Great Expectations v3 API.
        - context (DataContext, optional): an in-memory GE DataContext object. e.g.
            `ge.data_context.DataContext()` If not provided then `context_root_dir` will be used to
            look for one.
        - assets_to_validate (list, optional): A list of assets to validate when running the
            validation operator. Only used in the Great Expectations v2 API
        - batch_kwargs (dict, optional): a dictionary of batch kwargs to be used when validating
            assets. Only used in the Great Expectations v2 API
        - expectation_suite_name (str, optional): the name of an expectation suite to be used when
            validating assets. Only used in the Great Expectations v2 API
        - context_root_dir (str, optional): the absolute or relative path to the directory holding
            your `great_expectations.yml`
        - runtime_environment (dict, optional): a dictionary of great expectation config key-value
            pairs to overwrite your config in `great_expectations.yml`
        - run_name (str, optional): the name of this  Great Expectation validation run; defaults to
            the task slug
        - run_info_at_end (bool, optional): add run info to the end of the artifact generated by this
            task. Defaults to `True`.
        - disable_markdown_artifact (bool, optional): toggle the posting of a markdown artifact from
            this tasks. Defaults to `False`.
        - validation_operator (str, optional): configure the actions to be executed after running
            validation. Defaults to `action_list_operator`
        - evaluation_parameters (Optional[dict], optional): the evaluation parameters to use when
            running validation. For more information, see
            [example](https://docs.prefect.io/api/latest/tasks/great_expectations.html#rungreatexpectationsvalidation)
            and
            [docs](https://docs.greatexpectations.io/en/latest/reference/core_concepts/evaluation_parameters.html).
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task constructor
    """

    def __init__(
        self,
        checkpoint_name: str = None,
        ge_checkpoint: Checkpoint = None,
        checkpoint_kwargs: dict = None,
        context: ge.DataContext = None,
        assets_to_validate: list = None,
        batch_kwargs: dict = None,
        expectation_suite_name: str = None,
        context_root_dir: str = None,
        runtime_environment: Optional[dict] = None,
        run_name: str = None,
        run_info_at_end: bool = True,
        disable_markdown_artifact: bool = False,
        validation_operator: str = "action_list_operator",
        evaluation_parameters: Optional[dict] = None,
        **kwargs,
    ):
        self.checkpoint_name = checkpoint_name
        self.ge_checkpoint = ge_checkpoint
        self.checkpoint_kwargs = checkpoint_kwargs
        self.context = context
        self.assets_to_validate = assets_to_validate
        self.batch_kwargs = batch_kwargs
        self.expectation_suite_name = expectation_suite_name
        self.context_root_dir = context_root_dir
        self.runtime_environment = runtime_environment or dict()
        self.run_name = run_name
        self.run_info_at_end = run_info_at_end
        self.disable_markdown_artifact = disable_markdown_artifact
        self.validation_operator = validation_operator
        self.evaluation_parameters = evaluation_parameters

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "checkpoint_name",
        "ge_checkpoint",
        "checkpoint_kwargs",
        "context",
        "assets_to_validate",
        "batch_kwargs",
        "expectation_suite_name",
        "context_root_dir",
        "runtime_environment",
        "run_name",
        "run_info_at_end",
        "disable_markdown_artifact",
        "validation_operator",
        "evaluation_parameters",
    )
    def run(
        self,
        checkpoint_name: str = None,
        ge_checkpoint: Checkpoint = None,
        checkpoint_kwargs: dict = None,
        context: ge.DataContext = None,
        assets_to_validate: list = None,
        batch_kwargs: dict = None,
        expectation_suite_name: str = None,
        context_root_dir: str = None,
        runtime_environment: Optional[dict] = None,
        run_name: str = None,
        run_info_at_end: bool = True,
        disable_markdown_artifact: bool = False,
        validation_operator: str = "action_list_operator",
        evaluation_parameters: Optional[dict] = None,
    ):
        """
        Task run method.

        Args:
            - checkpoint_name (str, optional): the name of a pre-configured checkpoint; should match the
                filename of the checkpoint without the extension. Either checkpoint_name or
                checkpoint_config is required when using the Great Expectations v3 API.
            - ge_checkpoint (Checkpoint, optional): an in-memory GE `Checkpoint` object used to perform
                validation. If not provided then `checkpoint_name` will be used to load the specified
                checkpoint.
            - checkpoint_kwargs (Dict, optional): A dictionary whose keys match the parameters of
                `CheckpointConfig` which can be used to update and populate the task's Checkpoint at
                runtime.
            - context (DataContext, optional): an in-memory GE `DataContext` object. e.g.
                `ge.data_context.DataContext()` If not provided then `context_root_dir` will be used to
                look for one.
            - assets_to_validate (list, optional): A list of assets to validate when running the
                validation operator. Only used in the Great Expectations v2 API
            - batch_kwargs (dict, optional): a dictionary of batch kwargs to be used when validating
                assets. Only used in the Great Expectations v2 API
            - expectation_suite_name (str, optional): the name of an expectation suite to be used when
                validating assets. Only used in the Great Expectations v2 API
            - context_root_dir (str, optional): the absolute or relative path to the directory holding
                your `great_expectations.yml`
            - runtime_environment (dict, optional): a dictionary of great expectation config key-value
                pairs to overwrite your config in `great_expectations.yml`
            - run_name (str, optional): the name of this  Great Expectation validation run; defaults to
                the task slug
            - run_info_at_end (bool, optional): add run info to the end of the artifact generated by this
                task. Defaults to `True`.
            - disable_markdown_artifact (bool, optional): toggle the posting of a markdown artifact from
                this tasks. Defaults to `False`.
            - evaluation_parameters (Optional[dict], optional): the evaluation parameters to use when
                running validation. For more information, see
                [example](https://docs.prefect.io/api/latest/tasks/great_expectations.html#rungreatexpectationsvalidation)
                and
                [docs](https://docs.greatexpectations.io/en/latest/reference/core_concepts/evaluation_parameters.html).
            - validation_operator (str, optional): configure the actions to be executed after running
                validation. Defaults to `action_list_operator`.

        Raises:
            - 'signals.FAIL' if the validation was not a success

        Returns:
            - result
                ('great_expectations.validation_operators.types.validation_operator_result.ValidationOperatorResult'):
                The Great Expectations metadata returned from the validation if the v2 (batch_kwargs) API
                is used.

                ('great_expectations.checkpoint.checkpoint.CheckpointResult'):
                The Great Expectations metadata returned from running the provided checkpoint if a
                checkpoint name is provided.

        """

        if version.parse(ge.__version__) < version.parse("0.13.8"):
            self.logger.warning(
                f"You are using great_expectations version {ge.__version__} which may cause"
                "errors in this task. Please upgrade great_expections to 0.13.8 or later."
            )

        runtime_environment = runtime_environment or dict()
        checkpoint_kwargs = checkpoint_kwargs or dict()

        # Load context if not provided directly
        if not context:
            context = ge.DataContext(
                context_root_dir=context_root_dir,
                runtime_environment=runtime_environment,
            )

        # Check that the parameters are mutually exclusive
        if (
            sum(
                bool(x)
                for x in [
                    (expectation_suite_name and batch_kwargs),
                    assets_to_validate,
                    checkpoint_name,
                    ge_checkpoint,
                ]
            )
            != 1
        ):
            raise ValueError(
                "Exactly one of expectation_suite_name + batch_kwargs, assets_to_validate, "
                "checkpoint_name, or ge_checkpoint is required to run validation."
            )

        results = None
        # If there is a checkpoint or checkpoint name provided, run the checkpoint.
        # Checkpoints are the preferred deployment of validation configuration.
        if ge_checkpoint or checkpoint_name:
            ge_checkpoint = ge_checkpoint or context.get_checkpoint(checkpoint_name)
            results = ge_checkpoint.run(
                evaluation_parameters=evaluation_parameters,
                run_id=RunIdentifier(run_name or prefect.context.get("task_slug")),
                **checkpoint_kwargs,
            )
        else:
            # If assets are not provided directly through `assets_to_validate` then they need be loaded
            #   get batch from `batch_kwargs` and `expectation_suite_name`
            if not assets_to_validate:
                assets_to_validate = [
                    context.get_batch(batch_kwargs, expectation_suite_name)
                ]

            # Run validation operator
            results = context.run_validation_operator(
                validation_operator,
                assets_to_validate=assets_to_validate,
                run_id=RunIdentifier(run_name or prefect.context.get("task_slug")),
                evaluation_parameters=evaluation_parameters,
            )

        # Generate artifact markdown
        if not disable_markdown_artifact:
            validation_results_page_renderer = (
                ge.render.renderer.ValidationResultsPageRenderer(
                    run_info_at_end=run_info_at_end
                )
            )
            rendered_content_list = validation_results_page_renderer.render_validation_operator_result(
                # This also works with a CheckpointResult because of duck typing.
                # The passed in object needs a list_validation_results method that
                # returns a list of ExpectationSuiteValidationResult.
                validation_operator_result=results
            )
            markdown_artifact = " ".join(
                ge.render.view.DefaultMarkdownPageView().render(rendered_content_list)
            )

            create_markdown_artifact(markdown_artifact)

        if results.success is False:
            raise signals.FAIL(result=results)

        return results



"""
This module contains a collection of tasks to produce and consume Kafka events
"""

try:
    from prefect.tasks.kafka.kafka import KafkaBatchConsume, KafkaBatchProduce
except ImportError as err:
    raise ImportError(
        'Using `prefect.tasks.kafka` requires Prefect to be installed with the "kafka" extra.'
    ) from err

__all__ = ["KafkaBatchConsume", "KafkaBatchProduce"]



import psycopg2 as pg

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs


class PostgresExecute(Task):
    """
    Task for executing a query against a Postgres database.

    Args:
        - db_name (str): name of Postgres database
        - user (str): user name used to authenticate
        - host (str): database host address
        - port (int, optional): port used to connect to Postgres database, defaults to 5432 if
            not provided
        - query (str, optional): query to execute against database
        - data (tuple, optional): values to use in query, must be specified using placeholder
            in query string
        - commit (bool, optional): set to True to commit transaction, defaults to false
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        db_name: str = None,
        user: str = None,
        host: str = None,
        port: int = 5432,
        query: str = None,
        data: tuple = None,
        commit: bool = False,
        **kwargs
    ):
        self.db_name = db_name
        self.user = user
        self.host = host
        self.port = port
        self.query = query
        self.data = data
        self.commit = commit
        super().__init__(**kwargs)

    @defaults_from_attrs("db_name", "user", "host", "port", "query", "data", "commit")
    def run(
        self,
        db_name: str = None,
        user: str = None,
        host: str = None,
        port: int = 5432,
        query: str = None,
        data: tuple = None,
        commit: bool = False,
        password: str = None,
    ):
        """
        Task run method. Executes a query against Postgres database.

        Args:
            - db_name (str): name of Postgres database
            - user (str): user name used to authenticate
            - host (str): database host address
            - port (int, optional): port used to connect to Postgres database, defaults to 5432 if
                not provided
            - query (str, optional): query to execute against database
            - data (tuple, optional): values to use in query, must be specified using
                placeholder in query string
            - commit (bool, optional): set to True to commit transaction, defaults to false
            - password (str): password used to authenticate; should be provided from a `Secret` task

        Returns:
            - None

        Raises:
            - ValueError: if query parameter is None or a blank string
            - DatabaseError: if exception occurs when executing the query
        """
        if not query:
            raise ValueError("A query string must be provided")

        # connect to database, open cursor
        # allow psycopg2 to pass through any exceptions raised
        conn = pg.connect(
            dbname=db_name,
            user=user,
            password=password,
            host=host,
            port=port,
        )

        # try to execute query
        # context manager automatically rolls back failed transactions
        try:
            with conn, conn.cursor() as cursor:
                executed = cursor.execute(query=query, vars=data)
                if commit:
                    conn.commit()
                else:
                    conn.rollback()

            return executed

        # ensure connection is closed
        finally:
            conn.close()


class PostgresExecuteMany(Task):
    """
    Task for executing many queries against a Postgres database.

    Args:
        - db_name (str): name of Postgres database
        - user (str): user name used to authenticate
        - host (str): database host address
        - port (int, optional): port used to connect to Postgres database, defaults to 5432 if
            not provided
        - query (str, optional): query to execute against database
            is query string
        - data (List[tuple], optional): list of values to use in query, must be specified using
            placeholder
        - commit (bool, optional): set to True to commit transaction, defaults to false
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        db_name: str = None,
        user: str = None,
        host: str = None,
        port: int = 5432,
        query: str = None,
        data: list = None,
        commit: bool = False,
        **kwargs
    ):
        self.db_name = db_name
        self.user = user
        self.host = host
        self.port = port
        self.query = query
        self.data = data
        self.commit = commit
        super().__init__(**kwargs)

    @defaults_from_attrs("db_name", "user", "host", "port", "query", "data", "commit")
    def run(
        self,
        db_name: str = None,
        user: str = None,
        host: str = None,
        port: int = 5432,
        query: str = None,
        data: list = None,
        commit: bool = False,
        password: str = None,
    ):
        """
        Task run method. Executes many queries against Postgres database.

        Args:
            - db_name (str): name of Postgres database
            - user (str): user name used to authenticate
            - host (str): database host address
            - port (int, optional): port used to connect to Postgres database, defaults to 5432 if
                not provided
            - query (str, optional): query to execute against database
            - data (List[tuple], optional): list of values to use in query, must be specified using
                placeholder
            - commit (bool, optional): set to True to commit transaction, defaults to false
            - password (str): password used to authenticate; should be provided from a `Secret` task

        Returns:
            - None

        Raises:
            - ValueError: if query parameter is None or a blank string
            - DatabaseError: if exception occurs when executing the query
        """
        if not query:
            raise ValueError("A query string must be provided")

        if not data:
            raise ValueError("A data list must be provided")

        # connect to database, open cursor
        # allow psycopg2 to pass through any exceptions raised
        conn = pg.connect(
            dbname=db_name,
            user=user,
            password=password,
            host=host,
            port=port,
        )

        # try to execute query
        # context manager automatically rolls back failed transactions
        try:
            with conn, conn.cursor() as cursor:
                executed = cursor.executemany(query=query, vars_list=data)
                if commit:
                    conn.commit()
                else:
                    conn.rollback()

            return executed

        # ensure connection is closed
        finally:
            conn.close()


class PostgresFetch(Task):
    """
    Task for fetching results of query from Postgres database.

    Args:
        - db_name (str): name of Postgres database
        - user (str): user name used to authenticate
        - host (str): database host address
        - port (int, optional): port used to connect to Postgres database, defaults to 5432 if
            not provided
        - fetch (str, optional): one of "one" "many" or "all", used to determine how many
            results to fetch from executed query
        - fetch_count (int, optional): if fetch = 'many', determines the number of results to
            fetch, defaults to 10
        - query (str, optional): query to execute against database
        - data (tuple, optional): values to use in query, must be specified using placeholder
            is query string
        - commit (bool, optional): set to True to commit transaction, defaults to false
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        db_name: str = None,
        user: str = None,
        host: str = None,
        port: int = 5432,
        fetch: str = "one",
        fetch_count: int = 10,
        query: str = None,
        data: tuple = None,
        commit: bool = False,
        **kwargs
    ):
        self.db_name = db_name
        self.user = user
        self.host = host
        self.port = port
        self.fetch = fetch
        self.fetch_count = fetch_count
        self.query = query
        self.data = data
        self.commit = commit
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "db_name",
        "user",
        "host",
        "port",
        "fetch",
        "fetch_count",
        "query",
        "data",
        "commit",
    )
    def run(
        self,
        db_name: str = None,
        user: str = None,
        host: str = None,
        port: int = 5432,
        fetch: str = "one",
        fetch_count: int = 10,
        query: str = None,
        data: tuple = None,
        commit: bool = False,
        password: str = None,
        col_names: bool = False,
    ):
        """
        Task run method. Executes a query against Postgres database and fetches results.

        Args:
            - db_name (str): name of Postgres database
            - user (str): user name used to authenticate
            - host (str): database host address
            - port (int, optional): port used to connect to Postgres database, defaults to 5432 if
                not provided
            - fetch (str, optional): one of "one" "many" or "all", used to determine how many
                results to fetch from executed query
            - fetch_count (int, optional): if fetch = 'many', determines the number of results
                to fetch, defaults to 10
            - query (str, optional): query to execute against database
            - data (tuple, optional): values to use in query, must be specified using
                placeholder in query string
            - commit (bool, optional): set to True to commit transaction, defaults to false
            - password (str): password used to authenticate; should be provided from a `Secret` task
            - col_names (bool, optional): set to True to add column names to records, defaults to False

        Returns:
            - records (tuple or list of tuples): records from provided query

        Raises:
            - ValueError: if query parameter is None or a blank string
            - DatabaseError: if exception occurs when executing the query
        """
        if not query:
            raise ValueError("A query string must be provided")

        if fetch not in {"one", "many", "all"}:
            raise ValueError(
                "The 'fetch' parameter must be one of the following - ('one', 'many', 'all')"
            )

        # connect to database, open cursor
        # allow psycopg2 to pass through any exceptions raised
        conn = pg.connect(
            dbname=db_name,
            user=user,
            password=password,
            host=host,
            port=port,
        )

        # try to execute query
        # context manager automatically rolls back failed transactions
        try:
            with conn, conn.cursor() as cursor:
                cursor.execute(query=query, vars=data)

                # fetch results
                if fetch == "all":
                    records = cursor.fetchall()
                elif fetch == "many":
                    records = cursor.fetchmany(fetch_count)
                else:
                    records = cursor.fetchone()

                if commit:
                    conn.commit()

                if col_names:
                    names_list = [
                        col_description[0] for col_description in cursor.description
                    ]
                    header = [tuple(col_name for col_name in names_list)]

                    if isinstance(records, tuple):
                        records = [records]

                    records = header + records

            return records

        # ensure connection is closed
        finally:
            conn.close()



from datetime import date
from prefect import Task
from prefect.engine.signals import FAIL
from prefect.utilities.tasks import defaults_from_attrs
from collections import defaultdict

import json
import os
import requests
from requests.auth import HTTPBasicAuth

from typing import Union, List


class MixpanelExportTask(Task):
    """
    Task for performing an export using the Mixpanel Export API.
    More info about the API can be found at https://developer.mixpanel.com/reference/raw-event-export.

    Args:
        - api_secret (str, optional): The API secret key to use to authenticate
            to Mixpanel. Can be provided also via env var.
        - api_secret_env_var (str, optional): The name of the env var that contains
            the API secret key to use to authenticate to Mixpanel.
            `api_secret` takes precedence over `api_secret_env_var`.
        - from_date (str, optional): Start date of the export request.
            If provided as a string, it should be in the format `YYYY-MM-DD`.
            Default value is `2011-07-10`. This date is inclusive.
        - to_date (str, optional): End date of the export request.
            If provided as a string, it should be in the format `YYYY-MM-DD`.
            Default value is `prefect.context.today`. This date is inclusive.
        - limit (int, optional): The max number of events to return.
        - event (str, list, optional): The event, or events, that you wish
            to get the data for.
        - where (str, optional): An expression to filter events by.
            More info on expression sequence structure can be found
            at https://developer.mixpanel.com/reference/segmentation-expressions.
        - parse_response (bool, optional): Whether to parse the response into a JSON object.
            Default value is `False`.
        - use_eu_server (bool, optional): Whether to use the Mixpanel EU server to retrieve data.
            More info at https://help.mixpanel.com/hc/en-us/articles/360039135652-Data-Residency-in-EU.
            Default is `False`.
        - group_events: Whether to group events with the same name.
            This is taken into account only if `parse_response` is True.
        - **kwargs (dict, optional): Additional keyword arguments to pass to the
            Task constructor.
    """

    def __init__(
        self,
        api_secret: str = None,
        api_secret_env_var: str = None,
        from_date: str = "2011-07-10",  # The earliest date Mixpanel allows
        to_date: str = date.today().strftime("%Y-%m-%d"),
        limit: int = None,
        event: Union[str, List[str]] = None,
        where: str = None,
        parse_response: bool = False,
        use_eu_server: bool = False,
        group_events: bool = False,
        **kwargs,
    ):
        self.api_secret = api_secret
        self.api_secret_env_var = api_secret_env_var
        self.from_date = from_date
        self.to_date = to_date
        self.limit = limit
        self.event = event
        self.where = where
        self.parse_response = parse_response
        self.use_eu_server = use_eu_server
        self.group_events = group_events
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "api_secret",
        "api_secret_env_var",
        "from_date",
        "to_date",
        "limit",
        "event",
        "where",
        "parse_response",
        "use_eu_server",
        "group_events",
    )
    def run(
        self,
        api_secret: str = None,
        api_secret_env_var: str = None,
        from_date: str = None,
        to_date: str = None,
        limit: int = None,
        event: Union[str, List[str]] = None,
        where: str = None,
        parse_response: bool = False,
        use_eu_server: bool = False,
        group_events: bool = False,
    ):
        """
        Task run method to request a data export from Mixpanel using the Export API.

        Args:
            - api_secret (str, optional): The API secret key to use to authenticate
                to Mixpanel. Can be provided also via env var.
            - api_secret_env_var (str, optional): The name of the env var that contains
                the API secret key to use to authenticate to Mixpanel.
                `api_secret` takes precedence over `api_secret_env_var`.
            - from_date (str, optional): Start date of the export request.
                If provided as a string, it should be in the format `YYYY-MM-DD`.
                Default value is `2011-07-10`. This date is inclusive.
            - to_date (str, optional): End date of the export request.
                If provided as a string, it should be in the format `YYYY-MM-DD`.
                Default value is `prefect.context.today`. This date is inclusive.
            - limit (int, optional): The max number of events to return.
            - event (str, list, optional): The event, or events, that you wish
                to get the data for.
            - where (str, optional): An expression to filter events by.
                More info on expression sequence structure can be found
                at https://developer.mixpanel.com/reference/segmentation-expressions.
            - parse_response (bool, optional): Whether to parse the response into a JSON object.
                Default value is `False`.
            - use_eu_server (bool, optional): Whether to use the Mixpanel EU server to retrieve data.
                More info at
                https://help.mixpanel.com/hc/en-us/articles/360039135652-Data-Residency-in-EU.
                Default is `False`.
            - group_events: Whether to group events with the same name.
                This is taken into account only if `parse_response is True`.

        Returns:
            - if `parse_response` is False, then returns a `str` response pulled
                from the Export API, (which is basically a JSONL string).
            - if `parse_response` is True and `group_events` is True, then returns a `dict` where
                each key contains homogeneous events.
            - if `parse_response` is True and `group_events` is False, then returns
                a `list` of JSON objects obtained by parsing the response.

        Raises:
            - `ValueError` if both `api_secret` and `api_secret_env_var` are missing.
            - `ValueError` if `api_secret` is missing and `api_secret_env_var` is not found.
            - `prefect.engine.signals.FAIL` if the Mixpanel API returns an error.

        """
        if not api_secret and not api_secret_env_var:
            raise ValueError("Missing both `api_secret` and `api_secret_env_var`.")
        elif not api_secret and api_secret_env_var not in os.environ:
            raise ValueError("Missing `api_secret` and `api_secret_env_var` not found.")

        secret = None
        if api_secret:
            self.logger.debug("Got secret from `api_secret`")
            secret = api_secret
        else:
            self.logger.debug(
                "Got secret from env var passed from `api_secret_env_var`"
            )
            secret = os.environ[api_secret_env_var]

        params = {"from_date": from_date, "to_date": to_date}

        if limit:
            params["limit"] = limit

        if event:
            params["event"] = json.dumps([event] if isinstance(event, str) else event)

        if where:
            params["where"] = where

        url = "https://{server}.mixpanel.com/api/2.0/export".format(
            server="data-eu" if use_eu_server else "data"
        )

        response = requests.get(
            url=url,
            auth=HTTPBasicAuth(secret, ""),
            headers={"Accept": "application/json"},
            params=params,
        )

        if response.status_code != 200:
            msg = f"""
            Mixpanel export API error.
            Status code: {response.status_code}
            Reason: {response.reason}
            Text: {response.text}
            """
            raise FAIL(message=msg)

        events = response.text

        if not events:
            return None
        elif parse_response:
            received_events = [json.loads(event) for event in events.splitlines()]
            if group_events:
                grouped_events = defaultdict(list)
                for received_event in received_events:
                    grouped_events[received_event["event"]].append(
                        received_event["properties"]
                    )
                return dict(grouped_events)
            return received_events
        else:
            return events
